import pytest

from yt_idv.shader_objects import component_shaders, get_shader_combos


@pytest.mark.parametrize("render_name", component_shaders.keys())
def test_get_shader_combos_all_components(render_name):
    # default call to get_shader_combos should return all shaders
    expected = component_shaders[render_name]
    actual = get_shader_combos(render_name)
    assert set(expected) == set(actual)


def test_get_shader_combos_coord():
    # only block_rendering supports non-cartesian for now, check that the supported
    # shaders are returned as expected.
    shaders = get_shader_combos("block_rendering", coord_system="spherical")
    not_supported = ["isocontours", "slice"]
    assert all(s not in shaders for s in not_supported)
    supported = ["max_intensity", "projection", "transfer_function", "constant"]
    for s in supported:
        assert s in shaders
    assert all(s in shaders for s in supported)



import argparse

import numpy as np
import yt

import yt_idv

# yt reminder: phi is the azimuthal angle (0 to 2pi)
# theta is the co-latitude, the angle from north (0 to pi)
# coord ordering here will be r, phi, theta

ax_order = ("r", "phi", "theta")

bbox_options = {
    "partial": np.array([[0.5, 1.0], [0.0, np.pi / 3], [np.pi / 4, np.pi / 2]]),
    "whole": np.array([[0.0, 1.0], [0.0, 2 * np.pi], [0, np.pi]]),
    "shell": np.array([[0.5, 1.0], [0.0, 2 * np.pi], [0, np.pi]]),
    "north_hemi": np.array([[0.0, 1.0], [0.0, 2 * np.pi], [0, 0.5 * np.pi]]),
    "north_shell": np.array([[0.5, 1.0], [0.0, 2 * np.pi], [0, 0.5 * np.pi]]),
    "south_hemi": np.array([[0.0, 1.0], [0.0, 2 * np.pi], [0.5 * np.pi, np.pi]]),
    "ew_hemi": np.array([[0.0, 1.0], [0.0, np.pi], [0.0, np.pi]]),
}

max_levs = {
    "partial": 4,
    "whole": 5,
    "shell": 4,
    "north_hemi": 5,
    "north_shell": 2,
    "south_hemi": 5,
    "ew_hemi": 5,
}


def _build_ds(bbox_key):
    bbox = bbox_options[bbox_key]

    bbox_wid = bbox[:, 1] - bbox[:, 0]
    bbox_c = np.mean(bbox, axis=1)
    sz_0 = np.array((64, 64, 64))
    max_lev = max_levs[bbox_key]

    # divide grid by 2 every time
    dd0 = bbox_wid / sz_0
    sz_i = sz_0.copy()
    grids = []
    for lev in range(max_lev):

        box_wid_factor = 2.0 * int(lev > 0) + int(lev == 0) * 1.0
        bbox_wid = bbox_wid / box_wid_factor
        le_i = bbox_c - bbox_wid / 2.0
        re_i = bbox_c + bbox_wid / 2.0

        # find closest start/end index in lev 0 grid
        start_i = np.round(le_i / dd0).astype(int)
        end_i = np.round(re_i / dd0).astype(int)
        sz_0 = end_i - start_i

        # recompute for rounding errors
        le_i = start_i * dd0
        re_i = le_i + sz_0 * dd0

        sz_i = sz_0 * 2**lev

        levp1 = np.full(sz_i, lev + 1.0)
        grid = {
            "left_edge": le_i,
            "right_edge": re_i,
            "dimensions": sz_i,
            "level": lev,
            ("stream", "density"): np.random.random(sz_i) * (lev + 1),
            ("stream", "lev_p1"): levp1,
        }
        grids.append(grid)

    ds = yt.load_amr_grids(
        grids,
        sz_0,
        bbox=bbox,
        length_unit=1,
        geometry="spherical",
        axis_order=ax_order,
    )

    return ds


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog="spherical_amr_rendering_with_refinement",
        description="Loads an example spherical dataset with grid refinement in yt_idv",
    )

    msg = f"The geometry subset to generate: one of {list(bbox_options.keys())}"
    parser.add_argument("-d", "--domain", default="partial", help=msg)

    msg = (
        "The field to plot. Provide a comma-separated string with field_type,field "
        "e.g., to plot the field tuple ('index', 'phi'): \n "
        "    $ python amr_spherical_volume_rendering.py -f index,x "
        "\nIf a single string is provided, a field type of gas is assumed."
    )
    parser.add_argument("-f", "--field", default="stream,density", help=msg)

    parser.add_argument(
        "--listfields", action=argparse.BooleanOptionalAction, default=True
    )

    args = parser.parse_args()

    if args.domain not in bbox_options:
        raise RuntimeError(
            f"domain must be one of {list(bbox_options.keys())}, found {args.domain}"
        )

    ds = _build_ds(args.domain)

    if args.listfields:
        print("available fields:")
        print(ds.field_list)

    fld = tuple(str(args.field).split(","))

    rc = yt_idv.render_context(height=800, width=800, gui=True)
    sg = rc.add_scene(ds, fld, no_ghost=True)
    rc.scene.components[0].cmap_log = False
    rc.scene.components[0].sample_factor = 5.0

    rc.run()



"""
Example demonstrating manual construction of a SceneGraph with a BlockCollection
"""

import yt

import yt_idv
from yt_idv.cameras.trackball_camera import TrackballCamera
from yt_idv.scene_components.blocks import BlockRendering
from yt_idv.scene_data.block_collection import BlockCollection
from yt_idv.scene_graph import SceneGraph

ds = yt.load_sample("IsolatedGalaxy")
rc = yt_idv.render_context(height=800, width=800, gui=True)

c = TrackballCamera.from_dataset(ds)
rc.scene = SceneGraph(camera=c)
block_coll = BlockCollection(data_source=ds.all_data())
block_coll.add_data(("gas", "density"), no_ghost=True)
rc.scene.data_objects.append(block_coll)
rc.scene.components.append(BlockRendering(data=block_coll))

rc.run()



"""
Sometimes you want more control over data normalization and color-mapping. This
example demonstrates how to do that with BlockCollection data.
"""

import numpy as np
import yt

import yt_idv
from yt_idv.cameras.trackball_camera import TrackballCamera
from yt_idv.scene_components.blocks import BlockRendering
from yt_idv.scene_data.block_collection import BlockCollection
from yt_idv.scene_graph import SceneGraph

# create some data
shp = (32, 32, 32)
data = {"constant_field": np.full(shape=shp, fill_value=5.0)}
ds = yt.load_uniform_grid(data, shp, length_unit=1)


rc = yt_idv.render_context(height=800, width=800, gui=True)
c = TrackballCamera.from_dataset(ds)
rc.scene = SceneGraph(camera=c)

# the following BlockCollection initialization does the following:
#  1. sets the min, max values to be used for data normalization when
#     loading textures
#  2. specifies that those min, max vals should not be re-computed
#     (compute_min_max=False)
#  3. specifies that the data should always be normalized -- usually
#     normalization is skipped for constant fields.
block_coll = BlockCollection(
    data_source=ds.all_data(),
    min_val=0.0,
    max_val=10.0,
    compute_min_max=False,
    always_normalize=True,
)
block_coll.add_data(("stream", "constant_field"), no_ghost=True)
rc.scene.data_objects.append(block_coll)

# set fixed cmap ranges. Data are in normalized space. A constant data value of 5.
# (the fill value used above) with [min val, max val] of [0., 10.] normalizes to 0.5.
# So setting the fixed cmap min and max to 0 and 1 will result in a constant color
# at the middle of whatever colormap is selected.
block_rendering = BlockRendering(
    data=block_coll,
    fixed_cmap_min=0.0,
    fixed_cmap_max=1.0,
    cmap_log=False,
)

rc.scene.components.append(block_rendering)

rc.run()



import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.scene_annotations.base_annotation import SceneAnnotation
from yt_idv.scene_data.block_collection import BlockCollection


class BlockOutline(SceneAnnotation):
    """
    A class that renders outlines of block data.
    """

    name = "block_outline"
    data = traitlets.Instance(BlockCollection)
    box_width = traitlets.CFloat(0.1)
    box_color = traitlets.Tuple((1.0, 1.0, 1.0), trait=traitlets.CFloat())
    box_alpha = traitlets.CFloat(1.0)

    def draw(self, scene, program):
        GL.glDisable(GL.GL_CULL_FACE)
        each = self.data.vertex_array.each
        for tex_ind, _tex, _bitmap_tex in self.data.viewpoint_iter(scene.camera):
            GL.glDrawArrays(GL.GL_POINTS, tex_ind * each, each)

    def render_gui(self, imgui, renderer, scene):
        changed = super().render_gui(imgui, renderer, scene)
        _, bw = imgui.slider_float("Width", self.box_width, 0.001, 2.50)
        if _:
            self.box_width = bw
        changed = changed or _
        _, (r, g, b) = imgui.color_edit3("Color", *self.box_color)
        if _:
            self.box_color = (r, g, b)
        changed = changed or _
        _, ba = imgui.slider_float("Alpha", self.box_alpha, 0.0, 1.0)
        if _:
            self.box_alpha = ba
        changed = changed or _
        return changed

    def _set_uniforms(self, scene, shader_program):
        shader_program._set_uniform("box_width", self.box_width)
        shader_program._set_uniform("box_color", np.array(self.box_color))
        shader_program._set_uniform("box_alpha", self.box_alpha)



import yt

import yt_idv
from yt_idv.scene_components.octree_blocks import OctreeBlockRendering
from yt_idv.scene_data.octree_block_collection import OctreeBlockCollection

ds = yt.load_sample("output_00080")
dd = ds.all_data()

rc = yt_idv.render_context(height=800, width=800, gui=True)
sg = rc.add_scene(ds, None, no_ghost=True)

odata = OctreeBlockCollection(data_source=dd)
odata.add_data("density")
oren = OctreeBlockRendering(data=odata, display_name="density")

sg.data_objects.append(odata)
sg.components.append(oren)

rc.run()



import numpy as np
import traitlets

from yt_idv.opengl_support import VertexArray, VertexAttribute, compute_box_geometry
from yt_idv.scene_data.base_data import SceneData
from yt_idv.traitlets_support import YTPositionTrait


class BoxData(SceneData):
    name = "box_data"
    left_edge = YTPositionTrait([0.0, 0.0, 0.0])
    right_edge = YTPositionTrait([1.0, 1.0, 1.0])

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        va = VertexArray(name="box_outline", each=36)
        data = compute_box_geometry(self.left_edge, self.right_edge).copy()
        va.attributes.append(
            VertexAttribute(name="model_vertex", data=data.astype("f4"))
        )
        N = data.size // 4
        le = np.concatenate([[self.left_edge.copy()] for _ in range(N)])
        re = np.concatenate([[self.right_edge.copy()] for _ in range(N)])
        dds = self.right_edge - self.left_edge
        dds = np.concatenate([[dds.copy()] for _ in range(N)])
        va.attributes.append(VertexAttribute(name="in_left_edge", data=le.astype("f4")))
        va.attributes.append(
            VertexAttribute(name="in_right_edge", data=re.astype("f4"))
        )
        va.attributes.append(VertexAttribute(name="in_dx", data=dds.astype("f4")))
        return va



import numpy as np
import yt


class BaseContext:
    image_widget = None
    snap_count = 0

    def __init__(self, width, height, **kwargs):
        self.width = width
        self.height = height

    def add_scene(self, ds, field, no_ghost=True):
        """
        add a :class:`SceneGraph` to the current context for the supplied yt data.

        Parameters
        ----------
        ds:
            A yt Dataset or Data Object to use as a source
        field: Union[str, Tuple[str, str]]
            the field to render. can be either a string or a yt field tuple of
            the form ('field_type', 'field').
        no_ghost: bool
            if True, skip ghost zones to improve performance.

        Returns
        -------
        scene: :class:`SceneGraph`

        """
        from ..scene_graph import SceneGraph

        self.scene = SceneGraph.from_ds(ds, field, no_ghost=no_ghost)
        return self.scene

    def add_image(self, width=None, height=None):
        import ipywidgets

        width = width or self.width
        height = height or self.height
        # We will initialize it
        if self.scene is not None:
            value = yt.write_bitmap(self.scene.image[:, :, :3], None)
        else:
            value = yt.write_bitmap(np.zeros((width, height, 3)), None)
        self.image_widget = ipywidgets.Image(width=width, height=height, value=value)
        return self.image_widget

    def snap(self, template=r"snap_%04i.png"):
        yt.write_bitmap(self.scene.image[:, :, :3], template % self.snap_count)
        self.snap_count += 1



import numpy as np
import pytest
import yt

from yt_idv.cameras.trackball_camera import TrackballCamera
from yt_idv.scene_components.blocks import BlockRendering
from yt_idv.scene_data.block_collection import BlockCollection
from yt_idv.scene_graph import SceneGraph


@pytest.fixture()
def ds_yt_ugrid():
    # create some data
    shp = (8, 8, 8)
    data = {"constant_field": np.full(shape=shp, fill_value=5.0)}
    ds = yt.load_uniform_grid(data, shp, length_unit=1)
    return ds


def test_block_collection_normalization(osmesa_empty_rc, ds_yt_ugrid):

    block_coll: BlockCollection = BlockCollection(
        data_source=ds_yt_ugrid.all_data(),
        min_val=0.0,
        max_val=10.0,
        compute_min_max=False,
        always_normalize=True,
    )
    block_coll.add_data(("stream", "constant_field"), no_ghost=True)

    assert np.allclose(block_coll.texture_objects[0].data, 0.5)


def test_block_rendering_cmap_norms(osmesa_empty_rc, ds_yt_ugrid, image_store):

    block_coll: BlockCollection = BlockCollection(
        data_source=ds_yt_ugrid.all_data(),
        min_val=0.0,
        max_val=10.0,
        compute_min_max=False,
        always_normalize=True,
    )
    block_coll.add_data(("stream", "constant_field"), no_ghost=True)

    block_rendering: BlockRendering = BlockRendering(
        data=block_coll,
        fixed_cmap_min=0.2,
        fixed_cmap_max=0.8,
        cmap_log=False,
    )

    block_rendering._reset_cmap_bounds()

    assert block_rendering.cmap_min == 0.2
    assert block_rendering.cmap_max == 0.8

    block_rendering.fixed_cmap_min = 0.0
    block_rendering.fixed_cmap_max = 1.0
    block_rendering._reset_cmap_bounds()
    assert block_rendering.cmap_min == 0.0
    assert block_rendering.cmap_max == 1.0

    c = TrackballCamera.from_dataset(ds_yt_ugrid)
    osmesa_empty_rc.scene = SceneGraph(camera=c)
    osmesa_empty_rc.scene.data_objects.append(block_coll)
    osmesa_empty_rc.scene.components.append(block_rendering)

    image_store(osmesa_empty_rc)



cimport cython

import numpy as np

cimport numpy as np
from libc.math cimport INFINITY, M_PI, acos, atan2, cos, sin, sqrt


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef (np.float64_t, np.float64_t, np.float64_t) _spherical_to_cartesian(np.float64_t r,
                           np.float64_t theta,
                           np.float64_t phi) noexcept nogil:
        # transform a single point in spherical coords to cartesian
        # r : radius
        # theta: colatitude
        # phi: azimuthal (longitudinal) angle
        cdef np.float64_t x, y, xy, z

        if r == 0.0:
            return 0.0, 0.0, 0.0

        xy = r * sin(theta)
        x = xy * cos(phi)
        y = xy * sin(phi)
        z = r * cos(theta)
        return x, y, z


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
cdef (np.float64_t, np.float64_t, np.float64_t) _cartesian_to_spherical(np.float64_t x,
                           np.float64_t y,
                           np.float64_t z) noexcept nogil:
        # transform a single point in cartesian coords to spherical, returns
        # r : radius
        # theta: colatitude
        # phi: azimuthal angle in range (0, 2pi)
        cdef np.float64_t r, theta, phi
        r = sqrt(x*x + y*y + z*z)
        theta = acos(z / r)
        phi = atan2(y, x)
        # atan2 returns -pi to pi, adjust to (0, 2pi)
        if phi < 0:
            phi = phi + 2 * M_PI
        return r, theta, phi


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def cartesian_to_spherical(np.ndarray x,
                           np.ndarray y,
                           np.ndarray z):
        # transform an array of points in cartesian coords to spherical, returns
        # r : radius
        # theta: colatitude
        # phi: azimuthal angle in range (0, 2pi)
        cdef np.ndarray[np.float64_t, ndim=1] r1d, th1d, phi1d
        cdef np.ndarray[np.float64_t, ndim=1] x1d, y1d, z1d
        cdef int i, n_x, ndim
        cdef np.int64_t[:] shp

        ndim = x.ndim

        shp = np.zeros((ndim,), dtype=np.int64)
        for i in range(ndim):
            shp[i] = x.shape[i]

        x1d = x.reshape(-1)
        y1d = y.reshape(-1)
        z1d = z.reshape(-1)

        n_x = x1d.size
        r1d = np.zeros((n_x,), dtype=np.float64)
        th1d = np.zeros((n_x,), dtype=np.float64)
        phi1d = np.zeros((n_x,), dtype=np.float64)


        with nogil:
            for i in range(n_x):
                r1d[i], th1d[i], phi1d[i] = _cartesian_to_spherical(x1d[i], y1d[i], z1d[i])

        r = r1d.reshape(shp)
        theta = th1d.reshape(shp)
        phi = phi1d.reshape(shp)
        return r, theta, phi


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def spherical_to_cartesian(np.ndarray r,
                           np.ndarray theta,
                           np.ndarray phi):

        # transform an array of points in spherical coords to cartesian
        cdef np.ndarray[np.float64_t, ndim=1] r1d, th1d, phi1d
        cdef np.ndarray[np.float64_t, ndim=1] x1d, y1d, z1d

        cdef np.int64_t[:] shp
        cdef int i, n_r, ndim

        ndim = r.ndim

        shp = np.zeros((ndim,), dtype=np.int64)
        for i in range(ndim):
            shp[i] = r.shape[i]

        r1d = r.reshape(-1)
        th1d = theta.reshape(-1)
        phi1d = phi.reshape(-1)

        n_r = r1d.size
        x1d = np.zeros((n_r,), dtype=np.float64)
        y1d = np.zeros((n_r,), dtype=np.float64)
        z1d = np.zeros((n_r,), dtype=np.float64)


        with nogil:
            for i in range(n_r):
                x1d[i], y1d[i], z1d[i] = _spherical_to_cartesian(r1d[i], th1d[i], phi1d[i])

        x = x1d.reshape(shp)
        y = y1d.reshape(shp)
        z = z1d.reshape(shp)
        return x, y, z


cdef void _reduce_2_bboxes(np.float64_t[3] xyz_0,
                           np.float64_t[3] dxyz_0,
                           np.float64_t[3] xyz_1,
                           np.float64_t[3] dxyz_1,
                           np.float64_t[3] xyz,
                           np.float64_t[3] dxyz) noexcept nogil:

    # find the effective bounding box given 2 others
    cdef np.float64_t le_i, re_i
    cdef int idim
    cdef int ndim = 3

    for idim in range(ndim):
        le_i = fmin(xyz_0[idim] - dxyz_0[idim]/2.,
                    xyz_1[idim] - dxyz_1[idim]/2.)
        re_i = fmax(xyz_0[idim] + dxyz_0[idim]/2.,
                    xyz_1[idim] + dxyz_1[idim]/2.)
        xyz[idim] = (le_i + re_i)/2.
        dxyz[idim] = re_i - le_i


cdef class MixedCoordBBox:
    # abstract class for calculating cartesian bounding boxes
    # from non-cartesian grid elements.
    cdef int get_cartesian_bbox(self,
                                np.float64_t pos0,
                                np.float64_t pos1,
                                np.float64_t pos2,
                                np.float64_t dpos0,
                                np.float64_t dpos1,
                                np.float64_t dpos2,
                                np.float64_t xyz_i[3],
                                np.float64_t dxyz_i[3]
                                ) noexcept nogil:
        pass


cdef class SphericalMixedCoordBBox(MixedCoordBBox):
    # Cartesian bounding boxes of spherical grid elements
    cdef int get_cartesian_bbox(self,
                        np.float64_t pos0,  # r
                        np.float64_t pos1,  # theta
                        np.float64_t pos2,  # phi
                        np.float64_t dpos0, # r
                        np.float64_t dpos1, # theta
                        np.float64_t dpos2, # phi
                        np.float64_t[3] xyz_i,
                        np.float64_t[3] dxyz_i,
                        ) noexcept nogil:
        """
        Calculate the cartesian bounding box for a single spherical volume element.

        Parameters
        ----------
        pos0: float
            radius value at element center
        pos1: float
            co-latitude angle value at element center, in range (0, pi) (theta in yt)
        pos2: float
            azimuthal angle value at element center, in range (0, 2pi) (phi in yt)
        dpos0: float
            radial width of element
        dpos1: float
            co-latitude angular width of element
        pos2: float
            azimuthal anglular width of element
        xyz_i:
            3-element array to store the center of the resulting bounding box
        dxyz_i:
            3-element array to store the width of the resulting bounding box

        Returns
        -------
        int
            error code: 0 for success, 1 if something went wrong...

        Note: this function is recursive! If either angular width is above a cutoff (of
        0.2 radians or about 11 degrees), the element will be recursively divided.
        """

        cdef np.float64_t max_angle = 0.2  # about 11 degrees, if smaller, will subsample
        cdef np.float64_t dpos2_2, dpos1_2  # half-widths
        cdef np.float64_t pos2_2, pos1_2  # centers of half elements
        cdef np.float64_t[3] xyz_0, dxyz_0  # additional 1st sample point if needed
        cdef np.float64_t[3] xyz_1, dxyz_1  # additional 2nd sample point if needed

        if dpos1 < max_angle and dpos2 < max_angle:
            self._get_cartesian_bbox(
                            pos0, pos1,pos2,
                            dpos0, dpos1, dpos2,
                            xyz_i, dxyz_i
                            )
        elif dpos1 >= max_angle:
            # split into two, recursively get bbox, then get min/max of
            # the bboxes

            # first sample
            dpos1_2 = dpos1 / 2.0
            pos1_2 = pos1 - dpos1_2 / 2.
            self.get_cartesian_bbox(
                pos0, pos1_2, pos2, dpos0, dpos1_2, dpos2, xyz_0, dxyz_0
            )

            # second sample
            pos1_2 = pos1 + dpos1_2 / 2.
            self.get_cartesian_bbox(
                pos0, pos1_2, pos2, dpos0, dpos1_2, dpos2, xyz_1, dxyz_1
            )

            _reduce_2_bboxes(xyz_0, dxyz_0, xyz_1, dxyz_1, xyz_i, dxyz_i)

        elif dpos2 >= max_angle:
            # first sample
            dpos2_2 = dpos2 / 2.0
            pos2_2 = pos2 - dpos2_2 / 2.
            self.get_cartesian_bbox(
                pos0, pos1, pos2_2, dpos0, dpos1, dpos2_2, xyz_0, dxyz_0
            )

            # second sample
            pos2_2 = pos2 + dpos2_2 / 2.
            self.get_cartesian_bbox(
                pos0, pos1, pos2_2, dpos0, dpos1, dpos2_2, xyz_1, dxyz_1
            )

            _reduce_2_bboxes(xyz_0, dxyz_0, xyz_1, dxyz_1, xyz_i, dxyz_i)
        else:
            # this should be unreachable. Do not need to check for the case where
            # both dpos2 and dpos1 > max angle because of the recursive call!
            return 1
        return 0


    cdef void _get_cartesian_bbox(self,
                        np.float64_t pos0,
                        np.float64_t pos1,
                        np.float64_t pos2,
                        np.float64_t dpos0,
                        np.float64_t dpos1,
                        np.float64_t dpos2,
                        np.float64_t[3] xyz_i,
                        np.float64_t[3] dxyz_i
                        ) noexcept nogil:

        cdef np.float64_t r_i, theta_i, phi_i, dr_i, dtheta_i, dphi_i
        cdef np.float64_t h_dphi, h_dtheta, h_dr, r_r
        cdef np.float64_t xi, yi, zi, r_lr, theta_lr, phi_lr, phi_lr2, theta_lr2
        cdef np.float64_t xli, yli, zli, xri, yri, zri, r_xy, r_xy2
        cdef int isign_r, isign_ph, isign_th
        cdef np.float64_t sign_r, sign_th, sign_ph

        cdef np.float64_t M_PI_2 = M_PI / 2.0
        cdef np.float64_t M_PI_3_2 = 3. * M_PI / 2.0
        cdef np.float64_t NPY_2xPI = 2. * M_PI

        r_i = pos0
        theta_i = pos1
        phi_i = pos2
        dr_i = dpos0
        dtheta_i = dpos1
        dphi_i = dpos2

        # initialize the left/right values
        xli = INFINITY
        yli = INFINITY
        zli = INFINITY
        xri = -1.0 * INFINITY
        yri = -1.0 * INFINITY
        zri = -1.0 * INFINITY

        # find the min/max bounds over the 8 corners of the
        # spherical volume element.
        h_dphi =  dphi_i / 2.0
        h_dtheta =  dtheta_i / 2.0
        h_dr =  dr_i / 2.0
        for isign_r in range(2):
            for isign_ph in range(2):
                for isign_th in range(2):
                    sign_r = 1.0 - 2.0 * <float> isign_r
                    sign_th = 1.0 - 2.0 * <float> isign_th
                    sign_ph = 1.0 - 2.0 * <float> isign_ph
                    r_lr = r_i + sign_r * h_dr
                    theta_lr = theta_i + sign_th * h_dtheta
                    phi_lr = phi_i + sign_ph * h_dphi

                    xi, yi, zi = _spherical_to_cartesian(r_lr, theta_lr, phi_lr)

                    xli = fmin(xli, xi)
                    yli = fmin(yli, yi)
                    zli = fmin(zli, zi)
                    xri = fmax(xri, xi)
                    yri = fmax(yri, yi)
                    zri = fmax(zri, zi)

        # need to correct for special cases:
        # if polar angle, phi, spans pi/2, pi or 3pi/2 then just
        # taking the min/max of the corners will miss the cusp of the
        # element. When this condition is met, the x/y min/max will
        # equal +/- the projection of the max r in the xy plane -- in this case,
        # the theta angle that gives the max projection of r in
        # the x-y plane will change depending on the whether theta < or > pi/2,
        # so the following calculates for the min/max theta value of the element
        # and takes the max.
        # ALSO note, that the following does check for when an edge aligns with the
        # phi=0/2pi, it does not handle an element spanning the periodic boundary.
        # Oh and this may break down for large elements that span whole
        # quadrants...
        phi_lr =  phi_i - h_dphi
        phi_lr2 = phi_i + h_dphi
        theta_lr = theta_i - h_dtheta
        theta_lr2 = theta_i + h_dtheta
        r_r = r_i + h_dr
        if theta_lr < M_PI_2 and theta_lr2 > M_PI_2:
            r_xy = r_r
        else:
            r_xy = r_r * sin(theta_lr)
            r_xy2 = r_r * sin(theta_lr2)
            r_xy = fmax(r_xy, r_xy2)

        if phi_lr == 0.0 or phi_lr2 == NPY_2xPI:
            # need to re-check this, for when theta spans equator
            xri = r_xy
        elif phi_lr < M_PI_2 and phi_lr2  > M_PI_2:
            yri = r_xy
        elif phi_lr < M_PI and phi_lr2  > M_PI:
            xli = -r_xy
        elif phi_lr < M_PI_3_2 and phi_lr2  > M_PI_3_2:
            yli = -r_xy

        xyz_i[0] = (xri+xli)/2.
        xyz_i[1] = (yri+yli)/2.
        xyz_i[2] = (zri+zli)/2.
        dxyz_i[0] = xri-xli
        dxyz_i[1] = yri-yli
        dxyz_i[2] = zri-zli


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def cartesian_bboxes(MixedCoordBBox bbox_handler,
                      np.float64_t[:] pos0,
                      np.float64_t[:] pos1,
                      np.float64_t[:] pos2,
                      np.float64_t[:] dpos0,
                      np.float64_t[:] dpos1,
                      np.float64_t[:] dpos2,
                                      ):
    """
    Calculate the cartesian bounding boxes around non-cartesian volume elements

    Parameters
    ----------
    bbox_handler
        a MixedCoordBBox child instance
    pos0, pos1, pos2
        native coordinates of element centers
    dpos0, dpos1, dpos2
        element widths in native coordinates

    Returns
    -------
    x_y_z
        a 3-element tuples containing the cartesian bounding box centers
    dx_y_z
        a 3-element tuples containing the cartesian bounding box widths

    """

    cdef int i, n_pos, i_result
    cdef np.float64_t[3] xyz_i
    cdef np.float64_t[3] dxyz_i
    cdef int failure = 0
    cdef np.float64_t[:] x, y, z  # bbox centers
    cdef np.float64_t[:] dx, dy, dz  # bbox full widths

    n_pos = pos0.size
    x = np.empty_like(pos0)
    y = np.empty_like(pos0)
    z = np.empty_like(pos0)
    dx = np.empty_like(pos0)
    dy = np.empty_like(pos0)
    dz = np.empty_like(pos0)

    with nogil:
        for i in range(n_pos):

            i_result = bbox_handler.get_cartesian_bbox(pos0[i],
                                                    pos1[i],
                                                    pos2[i],
                                                    dpos0[i],
                                                    dpos1[i],
                                                    dpos2[i],
                                                    xyz_i,
                                                    dxyz_i)
            failure += i_result

            x[i] = xyz_i[0]
            y[i] = xyz_i[1]
            z[i] = xyz_i[2]
            dx[i] = dxyz_i[0]
            dy[i] = dxyz_i[1]
            dz[i] = dxyz_i[2]

    if failure > 0:
        raise RuntimeError("Unexpected error in get_cartesian_bbox.")

    x_y_z = (np.asarray(x), np.asarray(y), np.asarray(z))
    dx_y_z = (np.asarray(dx), np.asarray(dy), np.asarray(dz))
    return x_y_z, dx_y_z


@cython.cdivision(True)
@cython.boundscheck(False)
@cython.wraparound(False)
def cartesian_bboxes_edges(MixedCoordBBox bbox_handler,
                            np.float64_t[:] le0,
                            np.float64_t[:] le1,
                            np.float64_t[:] le2,
                            np.float64_t[:] re0,
                            np.float64_t[:] re1,
                            np.float64_t[:] re2,
                                      ):
    """
    Calculate the cartesian bounding boxes around non-cartesian volume elements

    Same as cartesian_bboxes, but supplying and returning element left/right edges.

    Parameters
    ----------
    bbox_handler
        a MixedCoordBBox child instance
    le0, le1, le2
        native coordinates of element left edges
    re0, re1, re2
        native coordinates of element right edges

    Returns
    -------
    xyz_le
        a 3-element tuples containing the cartesian bounding box left edges
    xyz_re
        a 3-element tuples containing the cartesian bounding box right edges

    """
    cdef int i, n_pos, i_result
    cdef np.float64_t[3] xyz_i
    cdef np.float64_t[3] dxyz_i
    cdef np.float64_t pos0, pos1, pos2, dpos0, dpos1, dpos2
    cdef int failure = 0
    cdef np.float64_t[:] x_le, y_le, z_le  # bbox left edges
    cdef np.float64_t[:] x_re, y_re, z_re  # bbox right edges

    n_pos = le0.size
    x_le = np.empty_like(le0)
    y_le = np.empty_like(le0)
    z_le = np.empty_like(le0)
    x_re = np.empty_like(le0)
    y_re = np.empty_like(le0)
    z_re = np.empty_like(le0)

    with nogil:
        for i in range(n_pos):
            pos0 = (le0[i] + re0[i]) / 2.
            pos1 = (le1[i] + re1[i]) / 2.
            pos2 = (le2[i] + re2[i]) / 2.
            dpos0 = re0[i] - le0[i]
            dpos1 = re1[i] - le1[i]
            dpos2 = re2[i] - le2[i]
            i_result = bbox_handler.get_cartesian_bbox(pos0,
                                                    pos1,
                                                    pos2,
                                                    dpos0,
                                                    dpos1,
                                                    dpos2,
                                                    xyz_i,
                                                    dxyz_i)
            failure += i_result

            x_le[i] = xyz_i[0] - dxyz_i[0] / 2.
            x_re[i] = xyz_i[0] + dxyz_i[0] / 2.
            y_le[i] = xyz_i[1] - dxyz_i[1] / 2.
            y_re[i] = xyz_i[1] + dxyz_i[1] / 2.
            z_le[i] = xyz_i[2] - dxyz_i[2] / 2.
            z_re[i] = xyz_i[2] + dxyz_i[2] / 2.


    if failure > 0:
        raise RuntimeError("Unexpected error in get_cartesian_bbox.")

    xyz_le = (np.asarray(x_le), np.asarray(y_le), np.asarray(z_le))
    xyz_re = (np.asarray(x_re), np.asarray(y_re), np.asarray(z_re))

    return xyz_le, xyz_re



import numpy as np
import traitlets
from OpenGL import GL

from yt_idv._cmyt_utilities import cmyt_names
from yt_idv.constants import FULLSCREEN_QUAD
from yt_idv.gui_support import add_popup_help
from yt_idv.opengl_support import (
    ColormapTexture,
    Framebuffer,
    VertexArray,
    VertexAttribute,
)
from yt_idv.scene_data.base_data import SceneData
from yt_idv.shader_objects import (
    PreprocessorDefinitionState,
    ShaderProgram,
    ShaderTrait,
    component_shaders,
    default_shader_combos,
)

_cmaps = ["arbre", "viridis", "magma", "doom", "cividis", "plasma", "RdBu", "coolwarm"]
_cmaps += [f"{_}_r" for _ in _cmaps]
# add in all the cmyt colormaps too! this will include the reversed maps too.
_cmaps += cmyt_names
_cmaps.sort(key=lambda v: v.lower())
_buffers = ["frame", "depth"]


_geom_directives = {"spherical": "SPHERICAL_GEOM"}


class SceneComponent(traitlets.HasTraits):
    data = traitlets.Instance(SceneData)
    base_quad = traitlets.Instance(SceneData)
    name = "undefined"
    priority = traitlets.CInt(0)
    visible = traitlets.Bool(True)
    use_db = traitlets.Bool(False)  # use depth buffer
    iso_tolerance = traitlets.CFloat(-1)  # the tolerance for finding isocontours
    iso_tol_is_pct = traitlets.Bool(False)  # if True, the tolerance is a fraction
    iso_log = traitlets.Bool(True)  # if True, iso values are base 10 exponents
    iso_layers = traitlets.List()  # the target values for isocontours
    iso_layers_alpha = traitlets.List()  # the transparency of isocontours
    display_bounds = traitlets.Tuple(
        traitlets.CFloat(),
        traitlets.CFloat(),
        traitlets.CFloat(),
        traitlets.CFloat(),
        default_value=(0.0, 1.0, 0.0, 1.0),
    )
    clear_region = traitlets.Bool(False)

    render_method = traitlets.Unicode(allow_none=True)
    fragment_shader = ShaderTrait(allow_none=True).tag(shader_type="fragment")
    geometry_shader = ShaderTrait(allow_none=True).tag(shader_type="geometry")
    vertex_shader = ShaderTrait(allow_none=True).tag(shader_type="vertex")
    fb = traitlets.Instance(Framebuffer)
    colormap_fragment = ShaderTrait(allow_none=True).tag(shader_type="fragment")
    colormap_vertex = ShaderTrait(allow_none=True).tag(shader_type="vertex")
    colormap = traitlets.Instance(ColormapTexture)
    _program1 = traitlets.Instance(ShaderProgram, allow_none=True)
    _program2 = traitlets.Instance(ShaderProgram, allow_none=True)
    _program1_pp_defs = traitlets.Instance(PreprocessorDefinitionState, allow_none=True)
    _program2_pp_defs = traitlets.Instance(PreprocessorDefinitionState, allow_none=True)
    _program1_invalid = True
    _program2_invalid = True
    _cmap_bounds_invalid = True
    _data_geometry = traitlets.Unicode(default_value="cartesian")

    display_name = traitlets.Unicode(allow_none=True)

    final_pass_vertex = ShaderTrait(allow_none=True).tag(shader_type="vertex")
    final_pass_fragment = ShaderTrait(allow_none=True).tag(shader_type="fragment")
    _final_pass = traitlets.Instance(ShaderProgram, allow_none=True)
    _final_pass_invalid = True

    # These attributes are just for colormap application
    fixed_cmap_min = traitlets.CFloat(None, allow_none=True)
    fixed_cmap_max = traitlets.CFloat(None, allow_none=True)
    cmap_min = traitlets.CFloat(None, allow_none=True)
    cmap_max = traitlets.CFloat(None, allow_none=True)
    cmap_log = traitlets.Bool(True)
    scale = traitlets.CFloat(1.0)

    # This attribute determines whether or not this component is "active"
    active = traitlets.Bool(True)

    @traitlets.observe("_data_geometry")
    def _update_geometry_pp_directives(self, change):
        if change["new"] != "cartesian" and change["old"] == "cartesian":
            geom_directive = _geom_directives[change["new"]]
            # this should only happen once on initial data load, so only handling the
            # change from the default value.
            current_shader = component_shaders[self.name][self.render_method]
            with self.hold_trait_notifications():
                for shd in ("vertex", "geometry", "fragment"):
                    # update the preprocessor directive state for this shader
                    self._program1_pp_defs.add_definition(shd, (geom_directive, ""))
                    self._program1_pp_defs.add_definition(
                        shd, ("NONCARTESIAN_GEOM", "")
                    )
                    pp_defs = self._program1_pp_defs[shd]

                    # update shader attributes of self
                    new_shader_tuple = (current_shader[f"first_{shd}"], pp_defs)
                    setattr(self, f"{shd}_shader", new_shader_tuple)

                    # all of the above shader attribute changes will trigger further
                    # traitlet changes that cause _recompile_shader to run

        self._data_geometry = change["new"]

    @traitlets.observe("data")
    def _update_geometry(self, change):
        if (
            hasattr(change["new"], "_yt_geom_str")
            and change["new"]._yt_geom_str == "spherical"
        ):
            self._data_geometry = change["new"]._yt_geom_str

    @traitlets.observe("display_bounds")
    def _change_display_bounds(self, change):
        # We need to update the framebuffer if the width or height has changed
        # Same thing is true if the total pixel size has changed, but that is
        # not doable from in here.
        if change["old"] == traitlets.Undefined:
            return
        old_width = change["old"][1] - change["old"][0]
        old_height = change["old"][3] - change["old"][2]
        new_width = change["new"][1] - change["new"][0]
        new_height = change["new"][3] - change["new"][2]
        if old_width != new_width or old_height != new_height:
            self.fb = Framebuffer()

    def render_gui(self, imgui, renderer, scene):
        changed, self.visible = imgui.checkbox("Visible", self.visible)
        _, self.use_db = imgui.checkbox("Depth Buffer", self.use_db)
        _ = add_popup_help(
            imgui, "If checked, will render the depth buffer of the current view."
        )
        changed = changed or _

        if imgui.button("Recompile Shader"):
            changed = self._recompile_shader()
        _, cmap_index = imgui.listbox(
            "Colormap", _cmaps.index(self.colormap.colormap_name), _cmaps
        )
        if _:
            self.colormap.colormap_name = _cmaps[cmap_index]
        changed = changed or _
        _ = add_popup_help(imgui, "Select the colormap to use for the rendering.")
        changed = changed or _
        _, self.cmap_log = imgui.checkbox("Take log", self.cmap_log)
        changed = changed or _
        _ = add_popup_help(
            imgui, "If checked, the rendering will use log-normalized values."
        )
        changed = changed or _
        if imgui.button("Reset Colorbounds"):
            self._cmap_bounds_invalid = True
            changed = True
        _ = add_popup_help(imgui, "Click to reset the colorbounds of the current view.")
        changed = changed or _

        if self.render_method == "isocontours":
            _ = self._render_isolayer_inputs(imgui)
            changed = changed or _

        return changed

    @traitlets.observe("iso_log")
    def _switch_iso_log(self, change):
        # if iso_log, then the user is setting 10**x, otherwise they are setting
        # x directly. So when toggling this checkbox we convert the existing
        # values between the two forms.
        if change["old"]:
            # if True, we were taking the log, but now are not:
            self.iso_tolerance = 10**self.iso_tolerance
            new_iso_layers = [10**iso_val for iso_val in self.iso_layers]
            self.iso_layers = new_iso_layers
        else:
            # we were not taking the log but now we are, so convert to the exponent
            self.iso_tolerance = np.log10(self.iso_tolerance)
            new_iso_layers = [np.log10(iso_val) for iso_val in self.iso_layers]
            self.iso_layers = new_iso_layers

    @traitlets.default("display_name")
    def _default_display_name(self):
        return self.name

    @traitlets.default("render_method")
    def _default_render_method(self):
        return default_shader_combos[self.name]

    @traitlets.default("_program1_pp_defs")
    def _default_program1_pp_defs(self):
        return PreprocessorDefinitionState()

    @traitlets.default("_program2_pp_defs")
    def _default_program2_pp_defs(self):
        return PreprocessorDefinitionState()

    @traitlets.observe("render_method")
    def _change_render_method(self, change):
        new_combo = component_shaders[self.name][change["new"]]
        with self.hold_trait_notifications():
            self.vertex_shader = (
                new_combo["first_vertex"],
                self._program1_pp_defs["vertex"],
            )
            self.fragment_shader = (
                new_combo["first_fragment"],
                self._program1_pp_defs["fragment"],
            )
            self.geometry_shader = (
                new_combo.get("first_geometry", None),
                self._program1_pp_defs["geometry"],
            )
            self.colormap_vertex = (
                new_combo["second_vertex"],
                self._program2_pp_defs["vertex"],
            )
            self.colormap_fragment = (
                new_combo["second_fragment"],
                self._program2_pp_defs["fragment"],
            )

    @traitlets.observe("render_method")
    def _add_initial_isolayer(self, change):
        # this adds an initial isocontour entry when the render method
        # switches to isocontours and if there are no layers yet.
        if change["new"] == "isocontours" and len(self.iso_layers) == 0:
            self.iso_layers.append(0.0)
            self.iso_layers_alpha.append(1.0)

    @traitlets.default("fb")
    def _fb_default(self):
        return Framebuffer()

    @traitlets.observe("fragment_shader")
    def _change_fragment(self, change):
        # Even if old/new are the same
        self._program1_invalid = True

    @traitlets.observe("vertex_shader")
    def _change_vertex(self, change):
        # Even if old/new are the same
        self._program1_invalid = True

    @traitlets.observe("geometry_shader")
    def _change_geometry(self, change):
        self._program1_invalid = True

    @traitlets.observe("colormap_vertex")
    def _change_colormap_vertex(self, change):
        # Even if old/new are the same
        self._program2_invalid = True

    @traitlets.observe("colormap_fragment")
    def _change_colormap_fragment(self, change):
        # Even if old/new are the same
        self._program2_invalid = True

    @traitlets.observe("use_db")
    def _toggle_depth_buffer(self, changed):
        # invalidate the colormap when the depth buffer selection changes
        self._cmap_bounds_invalid = True

        # update the preprocessor state: USE_DB only present in the second
        # program, only update that one.
        if changed["new"]:
            self._program2_pp_defs.add_definition("fragment", ("USE_DB", ""))
        else:
            self._program2_pp_defs.clear_definition("fragment", ("USE_DB", ""))

        # update the colormap fragment with current render method
        current_combo = component_shaders[self.name][self.render_method]
        pp_defs = self._program2_pp_defs["fragment"]
        self.colormap_fragment = current_combo["second_fragment"], pp_defs
        self._recompile_shader()

    @traitlets.default("colormap")
    def _default_colormap(self):
        cm = ColormapTexture()
        cm.colormap_name = "arbre"
        return cm

    @traitlets.default("vertex_shader")
    def _vertex_shader_default(self):
        return component_shaders[self.name][self.render_method]["first_vertex"]

    @traitlets.default("fragment_shader")
    def _fragment_shader_default(self):
        return component_shaders[self.name][self.render_method]["first_fragment"]

    @traitlets.default("geometry_shader")
    def _geometry_shader_default(self):
        _ = component_shaders[self.name][self.render_method]
        return _.get("first_geometry", None)

    @traitlets.default("colormap_vertex")
    def _colormap_vertex_default(self):
        return component_shaders[self.name][self.render_method]["second_vertex"]

    @traitlets.default("colormap_fragment")
    def _colormap_fragment_default(self):
        return component_shaders[self.name][self.render_method]["second_fragment"]

    @traitlets.default("final_pass_vertex")
    def _final_pass_vertex_default(self):
        return "passthrough"

    @traitlets.default("final_pass_fragment")
    def _final_pass_fragment_default(self):
        return "display_border"

    @traitlets.default("base_quad")
    def _default_base_quad(self):
        bq = SceneData(
            name="fullscreen_quad",
            vertex_array=VertexArray(name="tri", each=6),
        )
        fq = FULLSCREEN_QUAD.reshape((6, 3), order="C")
        bq.vertex_array.attributes.append(
            VertexAttribute(name="vertexPosition_modelspace", data=fq)
        )
        return bq

    @property
    def program1(self):
        if self._program1_invalid:
            if self._program1 is not None:
                self._program1.delete_program()
            self._fragment_shader_default()
            self._program1 = ShaderProgram(
                self.vertex_shader,
                self.fragment_shader,
                self.geometry_shader,
            )
            self._program1_invalid = False
        return self._program1

    @property
    def program2(self):
        if self._program2_invalid:
            if self._program2 is not None:
                self._program2.delete_program()
            # The vertex shader will always be the same.
            # The fragment shader will change based on whether we are
            # colormapping or not.
            self._program2 = ShaderProgram(
                self.colormap_vertex,
                self.colormap_fragment,
            )
            self._program2_invalid = False
        return self._program2

    @property
    def final_pass(self):
        if self._final_pass_invalid:
            if self._final_pass is not None:
                self._final_pass.delete_program()
            self._final_pass = ShaderProgram(
                self.final_pass_vertex, self.final_pass_fragment
            )
        return self._final_pass

    def _set_iso_uniforms(self, p):
        # these could be handled better by watching traits.
        p._set_uniform("iso_num_layers", int(len(self.iso_layers)))
        isolayervals = self._get_sanitized_iso_layers()
        p._set_uniform("iso_layers", isolayervals)
        p._set_uniform("iso_layer_tol", self._get_sanitized_iso_tol())
        avals = np.zeros((32,), dtype="float32")
        avals[: len(self.iso_layers)] = np.array(self.iso_layers_alpha)
        p._set_uniform("iso_alphas", avals)
        p._set_uniform("iso_min", float(self.data.min_val))
        p._set_uniform("iso_max", float(self.data.max_val))

    def run_program(self, scene):
        # Store this info, because we need to render into a framebuffer that is the
        # right size.
        if self.display_bounds != (0.0, 1.0, 0.0, 1.0):
            draw_boundary = 0.002
        else:
            draw_boundary = 0.0
        x0, y0, w, h = GL.glGetIntegerv(GL.GL_VIEWPORT)
        GL.glViewport(0, 0, w, h)
        if not self.visible:
            return
        with self.fb.bind(True):
            with self.program1.enable() as p:
                scene.camera._set_uniforms(scene, p)
                self._set_uniforms(scene, p)
                if self.render_method == "isocontours":
                    self._set_iso_uniforms(p)
                with self.data.vertex_array.bind(p):
                    self.draw(scene, p)

        if self._cmap_bounds_invalid:
            self._reset_cmap_bounds()

        with self.colormap.bind(0):
            with self.fb.input_bind(1, 2):
                with self.program2.enable() as p2:
                    with scene.bind_buffer():
                        p2._set_uniform("cmap", 0)
                        p2._set_uniform("fb_tex", 1)
                        p2._set_uniform("db_tex", 2)
                        # Note that we use cmap_min/cmap_max, not
                        # self.cmap_min/self.cmap_max.
                        p2._set_uniform("cmap_min", self.cmap_min)
                        p2._set_uniform("cmap_max", self.cmap_max)
                        p2._set_uniform("cmap_log", float(self.cmap_log))
                        with self.base_quad.vertex_array.bind(p2):
                            # Now we do our viewport globally, not just within
                            # the framebuffer
                            GL.glViewport(x0, y0, w, h)
                            GL.glDrawArrays(GL.GL_TRIANGLES, 0, 6)

        if draw_boundary > 0.0:
            with self.final_pass.enable() as p3:
                p3._set_uniform("draw_boundary", float(draw_boundary))
                if self.active:
                    boundary_color = np.array([0.0, 0.0, 1.0, 1.0], dtype="float32")
                else:
                    boundary_color = np.array([0.5, 0.5, 0.5, 1.0], dtype="float32")
                p3._set_uniform("boundary_color", boundary_color)
                with self.base_quad.vertex_array.bind(p3):
                    GL.glViewport(x0, y0, w, h)
                    GL.glDrawArrays(GL.GL_TRIANGLES, 0, 6)

    def draw(self, scene, program):
        raise NotImplementedError

    def _get_sanitized_iso_layers(self, normalize=True):
        # returns an array of the isocontour layer values, padded with 0s out
        # to max number of contours (32).
        iso_vals = np.asarray(self.iso_layers)
        if self.iso_log:
            iso_vals = 10**iso_vals

        if normalize:
            iso_vals = self.data._normalize_by_min_max(iso_vals)

        full_array = np.zeros(32, dtype="float32")
        full_array[: len(self.iso_layers)] = iso_vals
        return full_array

    def _get_sanitized_iso_tol(self):
        # isocontour selection conditions:
        #
        # absolute difference
        #   d - c <= eps
        # or percent difference
        #   (d - c) / c * 100 <= eps_pct
        #
        # where d is a raw data value, c is the target isocontour, eps
        # is an absolute difference, eps_f is a percent difference
        #
        # The data textures available on the shaders are normalized values:
        #   d_ = (d - min) / (max - min)
        # where max and min are the global min and max values across the entire
        # volume (e.g., over all blocks, not within a block)
        #
        # So in terms of normalized values, the absoulte difference condition
        # becomes
        #   d_ - c_ <= eps / (max - min)
        # where c_ is the target value normalized in the same way as d_.
        #
        # And the percent difference becomes
        #   (d_ - c_) * (max - min) / c * 100 <= eps_pct
        #       or
        #   d_ - c_ <= eps_pct / 100 * c / (max - min)
        # so that the allowed tolerance is a function of the raw target value
        # and so will vary with each layer.

        if self.iso_log:
            # the tol value is an exponent, convert
            tol = 10 ** float(self.iso_tolerance)
        else:
            tol = float(self.iso_tolerance)
        # always normalize tolerance
        tol = tol / self.data.val_range

        if self.iso_tol_is_pct:
            # tolerance depends on the layer value
            tol = tol * 0.01
            raw_layers = self._get_sanitized_iso_layers(normalize=False)
            final_tol = raw_layers * tol
        else:
            final_tol = np.full((32,), tol, dtype="float32")
        return final_tol

    def _recompile_shader(self) -> bool:
        # removes existing shaders, invalidates shader programs
        shaders = (
            "vertex_shader",
            "geometry_shader",
            "fragment_shader",
            "colormap_vertex",
            "colormap_fragment",
            "final_pass_vertex",
            "final_pass_fragment",
        )
        for shader_name in shaders:
            s = getattr(self, shader_name, None)
            if s:
                s.delete_shader()
        self._program1_invalid = self._program2_invalid = self._final_pass_invalid = (
            True
        )
        return True

    def _render_isolayer_inputs(self, imgui) -> bool:
        changed = False
        if imgui.tree_node("Isocontours"):
            _, self.iso_log = imgui.checkbox("set exponent", self.iso_log)
            _ = add_popup_help(
                imgui, "If checked, will treat isocontour values as base-10 exponents."
            )
            changed = changed or _

            imgui.columns(2, "iso_tol_cols", False)

            _, self.iso_tolerance = imgui.input_float(
                "tol",
                self.iso_tolerance,
                flags=imgui.INPUT_TEXT_ENTER_RETURNS_TRUE,
            )
            _ = add_popup_help(imgui, "The tolerance for selecting an isocontour.")
            changed = changed or _

            imgui.next_column()
            _, self.iso_tol_is_pct = imgui.checkbox("%", self.iso_tol_is_pct)
            _ = add_popup_help(imgui, "If checked, the tolerance is a percent.")
            changed = changed or _
            imgui.columns(1)

            if imgui.button("Add Layer"):
                if len(self.iso_layers) < 32:
                    changed = True
                    self.iso_layers.append(0.0)
                    self.iso_layers_alpha.append(1.0)
            _ = self._construct_isolayer_table(imgui)
            changed = changed or _
            imgui.tree_pop()
        return changed

    def _construct_isolayer_table(self, imgui) -> bool:
        imgui.columns(3, "iso_layers_cols", False)

        i = 0
        changed = False
        while i < len(self.iso_layers):
            _, self.iso_layers[i] = imgui.input_float(
                f"Layer {i + 1}",
                self.iso_layers[i],
                flags=imgui.INPUT_TEXT_ENTER_RETURNS_TRUE,
            )
            _ = add_popup_help(imgui, "The value of the isocontour layer.")
            changed = changed or _

            imgui.next_column()
            _, self.iso_layers_alpha[i] = imgui.input_float(
                f"alpha {i}",
                self.iso_layers_alpha[i],
                flags=imgui.INPUT_TEXT_ENTER_RETURNS_TRUE,
            )
            _ = add_popup_help(imgui, "The opacity of the isocontour layer.")
            changed = changed or _

            imgui.next_column()
            if imgui.button("Remove##rl" + str(i + 1)):
                self.iso_layers.pop(i)
                self.iso_layers_alpha.pop(i)
                i -= 1
                _ = True
            changed = changed or _
            imgui.next_column()
            i += 1
        imgui.columns(1)

        return changed

    def _reset_cmap_bounds(self, print_new_bounds=True):

        data = self.fb.data
        if self.use_db:
            data[:, :, :3] = self.fb.depth_data[:, :, None]
        data = data[data[:, :, 3] > 0][:, 0]
        if data.size > 0:
            self.cmap_min = data.min()
            self.cmap_max = data.max()

        if data.size == 0:
            self.cmap_min = 0.0
            self.cmap_max = 1.0

        # over-ride with the fixed values if they're set
        if self.fixed_cmap_max is not None:
            self.cmap_max = self.fixed_cmap_max

        if self.fixed_cmap_min is not None:
            self.cmap_min = self.fixed_cmap_min

        if print_new_bounds:
            print(f"Computed new cmap values {self.cmap_min} - {self.cmap_max}")

        self._cmap_bounds_invalid = False



import numpy as np
import traitlets
from traittypes import Array

from yt_idv.opengl_support import VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class CurveData(SceneData):
    """
    Data component for a single curve.
    """

    name = "curve_data"
    data = Array()
    n_vertices = traitlets.CInt()

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        va = VertexArray(name="vertices")
        return va

    def add_data(self, curve):
        """
        add curve data

        Parameters
        ----------
        curve: ArrayLike
            an array of shape (N, 3) specifying the position (in x, y, z) of
            N points.

        """
        # curve is a collection of ndarray of points
        assert curve.shape[0] > 1  # a curve needs at least 2 points
        assert curve.shape[1] == 3  # a curve needs at least 3 dimensions

        # add the singleton 4th dim
        data = np.ones((curve.shape[0], 4))
        data[:, 0:3] = curve

        self.n_vertices = curve.shape[0]
        self.data = data

        self.vertex_array.attributes.append(
            VertexAttribute(name="model_vertex", data=data.astype("f4"))
        )

        self.vertex_array.indices = np.arange(0, self.n_vertices).astype("uint32")
        self.size = self.n_vertices


class CurveCollection(CurveData):
    """Data component for a collection of curves"""

    name = "curve_collection"
    data = Array()
    n_vertices = traitlets.CInt()

    def __init__(self):
        self.n_curves = 0

    def add_curve(self, curve):
        """
        add a single curve to the collection

        Parameters
        ----------
        curve: ArrayLike
            an array of shape (N, 3) specifying the position (in x, y, z) of
            N points.

        """

        # curve is a collection of ndarray of points
        assert curve.shape[0] > 1  # a curve needs at least 2 points
        assert curve.shape[1] == 3  # a curve needs at least 3 dimensions

        # double up the indices to use GL_LINES
        index_range = np.arange(0, curve.shape[0])
        line_indices = np.column_stack([index_range, index_range]).ravel()[1:-1]
        data = curve[line_indices]
        data = np.column_stack([data, np.ones((data.shape[0],))])

        if self.data.shape:
            self.data = np.concatenate([self.data, data])
        else:
            self.data = data
        self.n_curves += 1

    def add_data(self):
        """
        finalize the current collection of curves.
        """
        self.n_vertices = self.data.shape[0]

        self.vertex_array.attributes.append(
            VertexAttribute(name="model_vertex", data=self.data.astype("f4"))
        )

        self.vertex_array.indices = np.arange(0, self.n_vertices).astype("uint32")
        self.size = self.n_vertices



from collections import defaultdict

import numpy as np
import traitlets
from yt.data_objects.data_containers import YTDataContainer

from yt_idv.opengl_support import Texture3D, VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class BlockCollection(SceneData):
    name = "block_collection"
    data_source = traitlets.Instance(YTDataContainer)
    texture_objects = traitlets.Dict(value_trait=traitlets.Instance(Texture3D))
    bitmap_objects = traitlets.Dict(value_trait=traitlets.Instance(Texture3D))
    blocks = traitlets.Dict(default_value=())
    scale = traitlets.Bool(False)
    blocks_by_grid = traitlets.Instance(defaultdict, (list,))
    grids_by_block = traitlets.Dict(default_value=())
    _yt_geom_str = traitlets.Unicode("cartesian")
    compute_min_max = traitlets.Bool(True)
    always_normalize = traitlets.Bool(False)

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        return VertexArray(name="block_info", each=1)

    def add_data(self, field, no_ghost=False):
        r"""Adds a source of data for the block collection.

        Given a `data_source` and a `field` to populate from, adds the data
        to the block collection so that is able to be rendered.

        Parameters
        ----------
        data_source : YTRegion
            A YTRegion object to use as a data source.
        field : string
            A field to populate from.
        no_ghost : bool (False)
            Should we speed things up by skipping ghost zone generation?
        """
        self.data_source.tiles.set_fields([field], [False], no_ghost=no_ghost)

        self._yt_geom_str = str(self.data_source.ds.geometry)
        # note: casting to string for compatibility with new and old geometry
        # attributes (now an enum member in latest yt),
        # see https://github.com/yt-project/yt/pull/4244

        # Every time we change our data source, we wipe all existing ones.
        # We now set up our vertices into our current data source.
        vert, dx, le, re = [], [], [], []

        min_val = +np.inf
        max_val = -np.inf
        if self.scale and self._yt_geom_str == "cartesian":
            left_min = np.ones(3, "f8") * np.inf
            right_max = np.ones(3, "f8") * -np.inf
            for block in self.data_source.tiles.traverse():
                np.minimum(left_min, block.LeftEdge, left_min)
                np.maximum(right_max, block.LeftEdge, right_max)
            scale = right_max.max() - left_min.min()
            for block in self.data_source.tiles.traverse():
                block.LeftEdge -= left_min
                block.LeftEdge /= scale
                block.RightEdge -= left_min
                block.RightEdge /= scale
        for i, block in enumerate(self.data_source.tiles.traverse()):
            min_val = min(min_val, np.nanmin(np.abs(block.my_data[0])).min())
            max_val = max(max_val, np.nanmax(np.abs(block.my_data[0])).max())
            self.blocks[id(block)] = (i, block)
            vert.append([1.0, 1.0, 1.0, 1.0])
            dds = (block.RightEdge - block.LeftEdge) / block.source_mask.shape
            dx.append(dds.tolist())
            le.append(block.LeftEdge.tolist())
            re.append(block.RightEdge.tolist())
        for g, node, (sl, _, gi) in self.data_source.tiles.slice_traverse():
            block = node.data
            self.blocks_by_grid[g.id - g._id_offset].append((id(block), gi))
            self.grids_by_block[id(node.data)] = (g.id - g._id_offset, sl)

        if self.compute_min_max:
            if hasattr(min_val, "in_units"):
                min_val = min_val.d
            if hasattr(max_val, "in_units"):
                max_val = max_val.d
            self.min_val = min_val
            self.max_val = max_val

        # Now we set up our buffer
        vert = np.array(vert, dtype="f4")
        dx = np.array(dx, dtype="f4")
        le = np.array(le)
        re = np.array(re)
        if self._yt_geom_str == "cartesian":
            # Note: the block LeftEdge and RightEdge arrays are plain np arrays in
            # units of code_length, so need to convert to unitary units (in range 0,1)
            # after the fact:
            units = self.data_source.ds.units
            ratio = (units.code_length / units.unitary).base_value
            dx = dx * ratio
            le = le * ratio
            re = re * ratio
            LE = np.array([b.LeftEdge for i, b in self.blocks.values()]).min(axis=0)
            RE = np.array([b.RightEdge for i, b in self.blocks.values()]).max(axis=0)
            self.diagonal = np.sqrt(((RE - LE) ** 2).sum())
        elif self._yt_geom_str == "spherical":
            rad_index = self.data_source.ds.coordinates.axis_id["r"]
            max_r = self.data_source.ds.domain_right_edge[rad_index]
            le[:, rad_index] = le[:, rad_index] / max_r
            re[:, rad_index] = re[:, rad_index] / max_r
            dx[:, rad_index] = dx[:, rad_index] / max_r

        self._set_geometry_attributes(le, re, dx)
        self.vertex_array.attributes.append(
            VertexAttribute(name="model_vertex", data=vert)
        )
        self.vertex_array.attributes.append(VertexAttribute(name="in_dx", data=dx))
        self.vertex_array.attributes.append(
            VertexAttribute(name="in_left_edge", data=le.astype("f4"))
        )
        self.vertex_array.attributes.append(
            VertexAttribute(name="in_right_edge", data=re.astype("f4"))
        )

        # Now we set up our textures
        self._load_textures()

    def _set_geometry_attributes(self, le, re, dx):
        # set any vertex_array attributes that depend on the yt geometry type
        #
        # for spherical coordinates, the radial component of le, re and dx
        # should already be normalized in the range of (0, 1)

        if self._yt_geom_str == "cartesian":
            return
        elif self._yt_geom_str == "spherical":
            from yt_idv.utilities.coordinate_utilities import (
                SphericalMixedCoordBBox,
                cartesian_bboxes_edges,
            )

            axis_id = self.data_source.ds.coordinates.axis_id

            # first, we need an approximation of the grid spacing
            # in cartesian coordinates. this is used by the
            # ray tracing engine to determine along-ray step size
            # so doesn't have to be exact. the ordering also
            # doesn't matter since it's the min value that will
            # influence step size. So here, we find some representative
            # lengths: the change in radius across the element,
            # the arc lengths of an element using average values of r
            # and theta where needed (average values avoid the edge case
            # of 0. values, which will cause the shader to crash)
            dr = dx[:, axis_id["r"]]
            rh = (le[:, axis_id["r"]] + re[:, axis_id["r"]]) / 2
            rdtheta = rh * dx[:, axis_id["theta"]]
            th = (le[:, axis_id["theta"]] + re[:, axis_id["theta"]]) / 2
            xy = rh * np.sin(th)
            rdphi = xy * dx[:, axis_id["phi"]]
            dx_cart = np.column_stack([dr, rdtheta, rdphi])

            # cartesian bbox calculations
            bbox_handler = SphericalMixedCoordBBox()
            le_cart, re_cart = cartesian_bboxes_edges(
                bbox_handler,
                le[:, axis_id["r"]],
                le[:, axis_id["theta"]],
                le[:, axis_id["phi"]],
                re[:, axis_id["r"]],
                re[:, axis_id["theta"]],
                re[:, axis_id["phi"]],
            )
            le_cart = np.column_stack(le_cart)
            re_cart = np.column_stack(re_cart)

            # cartesian le, re, width of whole domain
            domain_le = le_cart.min(axis=0)
            domain_re = re_cart.max(axis=0)
            domain_wid = domain_re - domain_le
            max_wid = np.max(domain_wid)

            # these will get passed down as uniforms to go from screen coords of
            # 0,1 to cartesian coords of domain_le to domain_re from which full
            # spherical coords can be calculated.
            self.cart_bbox_max_width = max_wid
            self.cart_bbox_le = domain_le
            self.cart_bbox_center = (domain_re + domain_le) / 2.0
            self.cart_min_dx = np.min(np.linalg.norm(dx_cart))

            self.vertex_array.attributes.append(
                VertexAttribute(name="le_cart", data=le_cart.astype("f4"))
            )
            self.vertex_array.attributes.append(
                VertexAttribute(name="re_cart", data=re_cart.astype("f4"))
            )
            self.vertex_array.attributes.append(
                VertexAttribute(name="dx_cart", data=dx_cart.astype("f4"))
            )

            # does not seem that diagonal is used anywhere, but recalculating to
            # be safe...
            self.diagonal = np.sqrt(((re_cart - le_cart) ** 2).sum())
        else:
            raise NotImplementedError(
                f"{self.name} does not implement {self._yt_geom_str} geometries."
            )

    def viewpoint_iter(self, camera):
        for block in self.data_source.tiles.traverse(viewpoint=camera.position):
            vbo_i, _ = self.blocks[id(block)]
            yield (vbo_i, self.texture_objects[vbo_i], self.bitmap_objects[vbo_i])

    def filter_callback(self, callback):
        # This is not efficient.  It calls it once for each node in a grid.
        # We do this the slow way because of the problem of ordering the way we
        # iterate over the grids and nodes.  This can be fixed at some point.
        for g_ind in self.blocks_by_grid:
            blocks = self.blocks_by_grid[g_ind]
            # Does this need an offset?
            grid = self.data_source.index.grids[g_ind]
            new_bitmap = callback(grid).astype("uint8")
            for b_id, _ in blocks:
                _, sl = self.grids_by_block[b_id]
                vbo_i, _ = self.blocks[b_id]
                self.bitmap_objects[vbo_i].data = new_bitmap[sl]

    def _load_textures(self):
        for block_id in sorted(self.blocks):
            vbo_i, block = self.blocks[block_id]
            n_data = np.abs(block.my_data[0]).copy(order="F").astype("float32").d
            # Avoid setting to NaNs
            if self.max_val != self.min_val or self.always_normalize:
                n_data = self._normalize_by_min_max(n_data)
                # blocks filled with identically 0 values will be
                # skipped by the shader, so offset by a tiny value.
                # see https://github.com/yt-project/yt_idv/issues/171
                n_data[n_data == 0.0] += np.finfo(np.float32).eps

            data_tex = Texture3D(data=n_data)
            bitmap_tex = Texture3D(
                data=block.source_mask * 255, min_filter="nearest", mag_filter="nearest"
            )
            self.texture_objects[vbo_i] = data_tex
            self.bitmap_objects[vbo_i] = bitmap_tex

    _grid_id_list = None

    @property
    def grid_id_list(self):
        """the 0-indexed grid ids that contain all the blocks"""
        if self._grid_id_list is None:
            gl = [gid for gid, _ in self.grids_by_block.values()]
            self._grid_id_list = np.unique(gl).tolist()
        return self._grid_id_list

    @property
    def intersected_grids(self):
        return [self.data_source.ds.index.grids[gid] for gid in self.grid_id_list]


def _block_collection_outlines(
    block_collection: BlockCollection,
    display_name: str = "block outlines",
    segments_per_edge: int = 20,
    outline_type: str = "blocks",
):
    """
    Build a CurveCollection and CurveCollectionRendering from BlockCollection
    bounding boxes for non-cartesian geometries.
    """

    if outline_type not in ("blocks", "grids"):
        msg = f"outline_type must be blocks or grids, found {outline_type}"
        raise ValueError(msg)

    if block_collection._yt_geom_str not in ("spherical",):
        msg = "_curves_from_block_data is not implemented for "
        msg += f"{block_collection._yt_geom_str} geometry."
        raise NotImplementedError(msg)

    from yt_idv.scene_components.curves import CurveCollectionRendering
    from yt_idv.scene_data.curve import CurveCollection

    data_collection = CurveCollection()

    if outline_type == "blocks":
        block_iterator = block_collection.data_source.tiles.traverse()
    else:
        # note this can be simplified after
        # https://github.com/yt-project/yt_idv/pull/179
        gids = [gid for gid, _ in block_collection.grids_by_block.values()]
        gids = np.unique(gids)
        ds = block_collection.data_source.ds
        block_iterator = [ds.index.grids[gid] for gid in gids]

    if block_collection._yt_geom_str == "spherical":
        from yt_idv.utilities.coordinate_utilities import spherical_to_cartesian

        # should move this down to cython to speed it up
        axis_id = block_collection.data_source.ds.coordinates.axis_id
        n_verts = segments_per_edge + 1

        rad_index = axis_id["r"]
        max_r = block_collection.data_source.ds.domain_right_edge[rad_index]

        for block in block_iterator:
            le_i = block.LeftEdge
            re_i = block.RightEdge

            r_min = le_i[axis_id["r"]] / max_r
            r_max = re_i[axis_id["r"]] / max_r

            theta_min = le_i[axis_id["theta"]]
            theta_max = re_i[axis_id["theta"]]

            phi_min = le_i[axis_id["phi"]]
            phi_max = re_i[axis_id["phi"]]

            theta_vals = np.linspace(theta_min, theta_max, n_verts)
            phi_vals = np.linspace(phi_min, phi_max, n_verts)

            # the r-variation will be straight lines always, only use 2 verts
            r_vals = np.linspace(r_min, r_max, 2)

            for r_val in (r_min, r_max):
                r = np.full(theta_vals.shape, r_val)
                for phi_val in (phi_min, phi_max):
                    phi = np.full(theta_vals.shape, phi_val)
                    x, y, z = spherical_to_cartesian(r, theta_vals, phi)
                    xyz = np.column_stack([x, y, z])
                    data_collection.add_curve(xyz)

                for theta_val in (theta_min, theta_max):
                    theta = np.full(phi_vals.shape, theta_val)
                    x, y, z = spherical_to_cartesian(r, theta, phi_vals)
                    xyz = np.column_stack([x, y, z])
                    data_collection.add_curve(xyz)

            for phi_val in (phi_min, phi_max):
                phi = np.full(r_vals.shape, phi_val)
                for theta_val in (theta_min, theta_max):
                    theta = np.full(r_vals.shape, theta_val)
                    x, y, z = spherical_to_cartesian(r_vals, theta, phi)
                    xyz = np.column_stack([x, y, z])
                    data_collection.add_curve(xyz)

    data_collection.add_data()  # call add_data() after done adding curves

    data_rendering = CurveCollectionRendering(data=data_collection)
    data_rendering.display_name = display_name

    return data_collection, data_rendering



import numpy as np

cimport libc.math as math
cimport numpy as np


def update_orientation(np.ndarray[np.float64_t, ndim=1] q1,
                        np.float64_t start_x, np.float64_t start_y,
                        np.float64_t end_x, np.float64_t end_y):
    # This is a Cython implementation of the TrackballCamera implementation of:
    #  1. map_to_surface(start_x, start_y)
    #  2. map_to_surface(end_x, end_y)
    #  3. Constructing w, x, y, z
    #  4. Normalizing this quaternion
    #  5. Multiplying by the existing orientation quaternion

    cdef np.float64_t w, x, y, z, q2[4]
    cdef np.float64_t old_map[3], new_map[3], mag
    cdef np.float64_t start_z, end_z
    cdef np.float64_t mag_start = (start_x * start_x + start_y * start_y)
    cdef np.float64_t mag_end = (end_x * end_x + end_y * end_y)

    if mag_start > 1.0:
        mag_start = mag_start**0.5
        start_x /= mag_start
        start_y /= mag_start
        start_z = 0.0
    else:
        start_z = (1.0 - mag_start)**0.5
    old_map[0] = start_x
    old_map[1] = -start_y
    old_map[2] = start_z

    if mag_end > 1.0:
        mag_end = mag_end**0.5
        end_x /= mag_end
        end_y /= mag_end
        end_z = 0.0
    else:
        end_z = (1.0 - mag_end)**0.5
    new_map[0] = end_x
    new_map[1] = -end_y
    new_map[2] = end_z
    w = old_map[0] * new_map[0] + old_map[1] * new_map[1] + old_map[2] * new_map[2]
    x = old_map[1] * new_map[2] - old_map[2] * new_map[1]
    y = old_map[2] * new_map[0] - old_map[0] * new_map[2]
    z = old_map[0] * new_map[1] - old_map[1] * new_map[0]

    mag = (w*w + x*x + y*y + z*z)**0.5

    q2[0] = w/mag
    q2[1] = x/mag
    q2[2] = y/mag
    q2[3] = z/mag

    # Now we multiply our quaternion

    w = q1[0]*q2[0] - q1[1]*q2[1] - q1[2]*q2[2] - q1[3]*q2[3]
    x = q1[0]*q2[1] + q1[1]*q2[0] + q1[2]*q2[3] - q1[3]*q2[2]
    y = q1[0]*q2[2] + q1[2]*q2[0] + q1[3]*q2[1] - q1[1]*q2[3]
    z = q1[0]*q2[3] + q1[3]*q2[0] + q1[1]*q2[2] - q1[2]*q2[1]

    q2[0] = w
    q2[1] = x
    q2[2] = y
    q2[3] = z
    return np.asarray(<np.float64_t[:4]> q2).copy()



import numpy as np
import traitlets

from yt_idv.opengl_support import VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class GridPositions(SceneData):
    name = "grid_positions"
    grid_list = traitlets.List()

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        va = VertexArray(name="grid_bounds")
        positions = []
        dx = []
        le = []
        re = []
        for g in self.grid_list:
            dx.append(g.dds.in_units("unitary").tolist())
            le.append(g.LeftEdge.in_units("unitary").tolist())
            re.append(g.RightEdge.in_units("unitary").tolist())
        positions = np.ones((len(self.grid_list), 4), dtype="f4")
        dx = np.array(dx).astype("f4")
        le = np.array(le).astype("f4")
        re = np.array(re).astype("f4")
        va.attributes.append(VertexAttribute(name="model_vertex", data=positions))
        va.attributes.append(VertexAttribute(name="in_left_edge", data=le))
        va.attributes.append(VertexAttribute(name="in_dx", data=dx))
        va.attributes.append(VertexAttribute(name="in_right_edge", data=re))
        return va



cimport numpy as np


cdef inline np.float64_t fmax(np.float64_t f0, np.float64_t f1) noexcept nogil:
    if f0 > f1: return f0
    return f1

cdef inline np.float64_t fmin(np.float64_t f0, np.float64_t f1) noexcept nogil:
    if f0 < f1: return f0
    return f1

cdef (np.float64_t, np.float64_t, np.float64_t) _spherical_to_cartesian(np.float64_t r,
                           np.float64_t theta,
                           np.float64_t phi) noexcept nogil


cdef (np.float64_t, np.float64_t, np.float64_t) _cartesian_to_spherical(np.float64_t x,
                           np.float64_t y,
                           np.float64_t z) noexcept nogil

cdef class MixedCoordBBox:
    cdef int get_cartesian_bbox(self,
                                np.float64_t pos0,
                                np.float64_t pos1,
                                np.float64_t pos2,
                                np.float64_t dpos0,
                                np.float64_t dpos1,
                                np.float64_t dpos2,
                                np.float64_t xyz_i[3],
                                np.float64_t dxyz_i[3]
                                ) noexcept nogil


cdef class SphericalMixedCoordBBox(MixedCoordBBox):

    cdef void _get_cartesian_bbox(
                    self,
                    np.float64_t pos0,
                    np.float64_t pos1,
                    np.float64_t pos2,
                    np.float64_t dpos0,
                    np.float64_t dpos1,
                    np.float64_t dpos2,
                    np.float64_t[3] xyz_i,
                    np.float64_t[3] dxyz_i
                    ) noexcept nogil



import numpy as np
import pytest
import yt

from yt_idv.cameras.trackball_camera import TrackballCamera
from yt_idv.scene_components.blocks import BlockRendering
from yt_idv.scene_components.curves import CurveCollectionRendering
from yt_idv.scene_data.block_collection import (
    BlockCollection,
    _block_collection_outlines,
)
from yt_idv.scene_data.curve import CurveCollection
from yt_idv.scene_graph import SceneGraph

bbox_options = {
    "partial": {
        "bbox": np.array([[0.5, 1.0], [0.0, np.pi / 3], [np.pi / 4, np.pi / 2]]),
        "field": ("index", "r"),
        "camera_position": [-0.5, -1, 2.5],
    },
    "whole": {
        "bbox": np.array([[0.0, 1.0], [0.0, 2 * np.pi], [0, np.pi]]),
        "field": ("index", "phi"),
    },
    "quadrant_shell": {
        "bbox": np.array([[0.6, 1.0], [0.0, np.pi / 2], [0.0, np.pi / 2]]),
        "field": ("index", "theta"),
    },
    "big_r": {
        "bbox": np.array([[0.0, 100.0], [0.0, 2 * np.pi], [0, np.pi]]),
        "field": ("index", "phi"),
    },
}


def _get_sph_yt_ds(bbox_option: str):
    sz = (32, 32, 32)
    fake_data = {"density": np.random.random(sz)}

    bbox = bbox_options[bbox_option]["bbox"]

    ds = yt.load_uniform_grid(
        fake_data,
        sz,
        bbox=bbox,
        nprocs=8,
        geometry="spherical",
        axis_order=("r", "phi", "theta"),
        length_unit="m",
    )
    return ds


@pytest.mark.parametrize("bbox_option", bbox_options.keys())
def test_spherical_bounds(osmesa_empty_rc, image_store, bbox_option):

    ds = _get_sph_yt_ds(bbox_option)
    dd = ds.all_data()

    field = bbox_options[bbox_option]["field"]
    osmesa_empty_rc.add_scene(dd, field, no_ghost=True)
    osmesa_empty_rc.scene.components[0].sample_factor = 5.0
    osmesa_empty_rc.scene.components[0].cmap_log = False
    cpos = bbox_options[bbox_option].get("camera_position", None)
    if cpos:
        osmesa_empty_rc.scene.camera.set_position(cpos)

    image_store(osmesa_empty_rc)


@pytest.mark.parametrize("nprocs", [1, 2, 4, 16])
def test_spherical_nprocs(osmesa_empty_rc, image_store, nprocs):

    bbox_option = "whole"
    ds = _get_sph_yt_ds(bbox_option)
    dd = ds.all_data()

    field = bbox_options[bbox_option]["field"]
    osmesa_empty_rc.add_scene(dd, field, no_ghost=True)
    osmesa_empty_rc.scene.components[0].sample_factor = 5.0
    osmesa_empty_rc.scene.components[0].cmap_log = False
    osmesa_empty_rc.scene.components[0]._reset_cmap_bounds()
    cpos = bbox_options[bbox_option].get("camera_position", None)
    if cpos:
        osmesa_empty_rc.scene.camera.set_position(cpos)

    image_store(osmesa_empty_rc)


@pytest.mark.parametrize("bbox_option", ["partial", "big_r"])
def test_block_collection_outlines(osmesa_empty_rc, image_store, bbox_option):

    ds = _get_sph_yt_ds(bbox_option)
    block_coll: BlockCollection = BlockCollection(
        data_source=ds.all_data(),
    )
    block_coll.add_data(("stream", "density"), no_ghost=True)

    block_rendering: BlockRendering = BlockRendering(
        data=block_coll,
        cmap_log=False,
    )

    curve_coll, curve_render = _block_collection_outlines(block_coll)

    assert isinstance(curve_coll, CurveCollection)
    assert isinstance(curve_render, CurveCollectionRendering)
    assert curve_coll.n_curves > 0

    c = TrackballCamera.from_dataset(ds)
    osmesa_empty_rc.scene = SceneGraph(camera=c)
    osmesa_empty_rc.scene.data_objects.append(block_coll)
    osmesa_empty_rc.scene.components.append(block_rendering)

    osmesa_empty_rc.scene.data_objects.append(curve_coll)
    osmesa_empty_rc.scene.components.append(curve_render)

    image_store(osmesa_empty_rc)



import yt

import yt_idv

ds = yt.load_sample("IsolatedGalaxy")
dd = ds.all_data()

rc = yt_idv.render_context("osmesa", width=1024, height=1024)
rc.add_scene(dd, "density", no_ghost=True)

image = rc.run()
yt.write_bitmap(image, "step1.png")

rc.scene.camera.move_forward(1.5)

image = rc.run()
yt.write_bitmap(image, "step2.png")

rc.scene.camera.set_position([1.0, 1.5, 3.0])
image = rc.run()
yt.write_bitmap(image, "step3_set_position.png")



import numpy as np
import traitlets
from OpenGL import GL

from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_data.curve import CurveCollection, CurveData


class CurveRendering(SceneComponent):
    """
    A class that renders user-specified curves.
    """

    name = "curve_rendering"
    data = traitlets.Instance(CurveData, help="The curve data.")
    curve_rgba = traitlets.Tuple((1.0, 1.0, 1.0, 1.0)).tag(trait=traitlets.CFloat())
    priority = 10

    def render_gui(self, imgui, renderer, scene):
        changed, self.visible = imgui.checkbox("Visible", self.visible)

        _, newRGBa = imgui.input_float4("RGBa", *self.curve_rgba)
        if _:
            self.curve_rgba = tuple(newRGBa)
        changed = changed or _

        return changed

    def draw(self, scene, program):
        GL.glEnable(GL.GL_CULL_FACE)
        GL.glCullFace(GL.GL_BACK)
        # note: line width != 1.0 is not supported on all platforms.
        # see https://github.com/yt-project/yt_idv/issues/174
        GL.glLineWidth(1.0)
        GL.glDrawArrays(GL.GL_LINE_STRIP, 0, self.data.n_vertices)

    def _set_uniforms(self, scene, shader_program):
        clr = np.array(self.curve_rgba).astype("f4")
        shader_program._set_uniform("curve_rgba", clr)


class CurveCollectionRendering(CurveRendering):
    """
    rendering a collection of curves
    """

    name = "multi_curve_rendering"
    curve_collection = traitlets.List()
    data = traitlets.Instance(CurveCollection)

    def draw(self, scene, program):
        GL.glEnable(GL.GL_CULL_FACE)
        GL.glCullFace(GL.GL_BACK)
        GL.glLineWidth(1.0)
        GL.glDrawArrays(GL.GL_LINES, 0, self.data.n_vertices)



"""Tests for `yt_idv` package."""

import numpy as np
import pytest
import yt
import yt.testing
from numpy.testing import assert_equal

import yt_idv
from yt_idv import shader_objects
from yt_idv.cameras.trackball_camera import TrackballCamera
from yt_idv.scene_components.blocks import BlockRendering
from yt_idv.scene_components.curves import CurveCollectionRendering, CurveRendering
from yt_idv.scene_data.block_collection import BlockCollection
from yt_idv.scene_data.curve import CurveCollection, CurveData
from yt_idv.scene_graph import SceneGraph


@pytest.fixture()
def osmesa_fake_amr():
    """Return an OSMesa context that has a "fake" AMR dataset added, with "radius"
    as the field.
    """
    ds = yt.testing.fake_amr_ds()
    dd = ds.all_data()
    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    rc.add_scene(dd, "radius", no_ghost=True)
    yield rc
    rc.osmesa.OSMesaDestroyContext(rc.context)


@pytest.fixture()
def osmesa_empty():
    """Return an OSMesa context that has no dataset."""
    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    ds = yt.testing.fake_amr_ds()
    rc.add_scene(ds, None)
    rc.ds = ds
    yield rc
    rc.osmesa.OSMesaDestroyContext(rc.context)


def test_snapshots(osmesa_fake_amr, image_store):
    """Check that we can make some snapshots."""
    osmesa_fake_amr.scene.components[0].render_method = "max_intensity"
    image_store(osmesa_fake_amr)
    osmesa_fake_amr.scene.components[0].render_method = "projection"
    image_store(osmesa_fake_amr)
    osmesa_fake_amr.scene.components[0].render_method = "transfer_function"
    image_store(osmesa_fake_amr)
    osmesa_fake_amr.scene.components[0]._recompile_shader()
    image_store(osmesa_fake_amr)


def test_camera_position(osmesa_fake_amr, image_store):
    """Check that we can update the camera position"""
    vm = osmesa_fake_amr.scene.camera.view_matrix
    osmesa_fake_amr.scene.camera.set_position([0.5, 2.0, 3.0])
    # check that the view matrix has changed
    assert np.sum(np.abs(vm - osmesa_fake_amr.scene.camera.view_matrix)) > 0.0
    image_store(osmesa_fake_amr)


def test_depth_buffer_toggle(osmesa_fake_amr, image_store):
    osmesa_fake_amr.scene.components[0].use_db = True
    image_store(osmesa_fake_amr)


def test_slice(osmesa_fake_amr, image_store):
    osmesa_fake_amr.scene.components[0].render_method = "slice"
    osmesa_fake_amr.scene.components[0].slice_position = (0.5, 0.5, 0.5)
    for ax in [0, 1, 2]:
        normal = [0.0, 0.0, 0.0]
        normal[ax] = 1.0
        osmesa_fake_amr.scene.components[0].slice_normal = tuple(normal)
        image_store(osmesa_fake_amr)
    osmesa_fake_amr.scene.components[0].slice_normal = (1.0, 1.0, 0.0)
    osmesa_fake_amr.scene.components[0].slice_position = (0.5, 0.25, 0.5)
    image_store(osmesa_fake_amr)


def test_annotate_boxes(osmesa_empty, image_store):
    """Check the box annotation."""
    osmesa_empty.scene.add_box([0.0, 0.0, 0.0], [1.0, 1.0, 1.0])
    image_store(osmesa_empty)
    osmesa_empty.scene.add_box([0.2, 0.2, 0.3], [0.8, 0.8, 0.7])
    image_store(osmesa_empty)
    osmesa_empty.scene.annotations[-1].box_width /= 2
    osmesa_empty.scene.annotations[-1].box_color = (1.0, 0.0, 0.0)
    image_store(osmesa_empty)


def test_annotate_grids(osmesa_empty, image_store):
    """Make sure we can add some grid positions."""
    from yt_idv.scene_annotations.grid_outlines import GridOutlines  # NOQA
    from yt_idv.scene_data.grid_positions import GridPositions  # NOQA

    gp = GridPositions(grid_list=osmesa_empty.ds.index.grids.tolist())
    osmesa_empty.scene.data_objects.append(gp)
    go = GridOutlines(data=gp)
    osmesa_empty.scene.components.append(go)
    image_store(osmesa_empty)
    osmesa_empty.scene.camera.offset_position(0.25)
    image_store(osmesa_empty)
    osmesa_empty.scene.camera.offset_position(0.5)
    image_store(osmesa_empty)


def test_annotate_text(osmesa_empty, image_store):
    """Test that text can be annotated and updated."""
    text = osmesa_empty.scene.add_text("Origin 0 0", origin=(0.0, 0.0))
    image_store(osmesa_empty)
    text.text = "Change text"
    image_store(osmesa_empty)
    text.text = "Origin -0.5 -0.5"
    text.origin = (-0.5, -0.5)
    image_store(osmesa_empty)
    text.origin = (0.0, 0.0)
    text.text = "S 1.0"
    image_store(osmesa_empty)
    text.text = "S 2.0"
    text.scale = 2.0
    image_store(osmesa_empty)


def test_isocontour_functionality(osmesa_fake_amr, image_store):
    osmesa_fake_amr.scene.components[0].render_method = "isocontours"
    image_store(osmesa_fake_amr)


def test_curves(osmesa_fake_amr, image_store):
    # add a single curve

    curved = CurveData()
    x1d = np.linspace(0, 1, 10)
    xyz = np.column_stack([x1d, x1d, np.zeros((10,))])
    curved.add_data(xyz)
    curve_render = CurveRendering(
        data=curved, curve_rgba=(1.0, 0.0, 0.0, 1.0), line_width=4
    )
    curve_render.display_name = "single streamline"
    osmesa_fake_amr.scene.data_objects.append(curved)
    osmesa_fake_amr.scene.components.append(curve_render)
    image_store(osmesa_fake_amr)

    curve_collection = CurveCollection()
    xyz = np.column_stack([x1d, np.zeros((10,)), x1d])
    curve_collection.add_curve(xyz)
    xyz = np.column_stack([np.zeros((10,)), x1d, x1d])
    curve_collection.add_curve(xyz)
    curve_collection.add_data()  # call add_data() after done adding curves

    cc_render = CurveCollectionRendering(
        data=curve_collection, curve_rgba=(0.2, 0.2, 0.2, 1.0), line_width=4
    )
    cc_render.display_name = "multiple streamlines"
    osmesa_fake_amr.scene.data_objects.append(curve_collection)
    osmesa_fake_amr.scene.components.append(cc_render)

    image_store(osmesa_fake_amr)


@pytest.fixture()
def set_very_bad_shader():
    # this temporarily points the default vertex shader source file to a
    # bad shader that will raise compilation errors.
    known_shaders = shader_objects.known_shaders
    good_shader = known_shaders["vertex"]["default"]["source"]
    known_shaders["vertex"]["default"]["source"] = "bad_shader.vert.glsl"
    yield known_shaders
    known_shaders["vertex"]["default"]["source"] = good_shader


def test_bad_shader(osmesa_empty, set_very_bad_shader):
    # this test is meant to check that a bad shader would indeed be caught
    # by the subsequent test_shader_programs test.
    shader_name = "box_outline"
    program = shader_objects.component_shaders[shader_name]["default"]

    vertex_shader = shader_objects._validate_shader(
        "vertex", program["first_vertex"], allow_null=False
    )
    fragment_shader = shader_objects._validate_shader(
        "fragment", program["first_fragment"], allow_null=False
    )
    with pytest.raises(RuntimeError, match="shader complilation error"):
        _ = shader_objects.ShaderProgram(vertex_shader, fragment_shader, None)


@pytest.mark.parametrize("shader_name", list(shader_objects.component_shaders.keys()))
def test_shader_programs(osmesa_empty, shader_name):
    for program in shader_objects.component_shaders[shader_name].values():

        vertex_shader = shader_objects._validate_shader(
            "vertex", program["first_vertex"], allow_null=False
        )
        assert isinstance(vertex_shader, shader_objects.Shader)
        fragment_shader = shader_objects._validate_shader(
            "fragment", program["first_fragment"], allow_null=False
        )
        assert isinstance(fragment_shader, shader_objects.Shader)
        geometry_shader = program.get("first_geometry", None)
        if geometry_shader is not None:
            geometry_shader = shader_objects._validate_shader(
                "geometry", geometry_shader, allow_null=False
            )
            assert isinstance(geometry_shader, shader_objects.Shader)

        _ = shader_objects.ShaderProgram(
            vertex_shader, fragment_shader, geometry_shader
        )

        colormap_vertex = shader_objects._validate_shader(
            "vertex", program["second_vertex"], allow_null=False
        )
        assert isinstance(colormap_vertex, shader_objects.Shader)
        colormap_fragment = shader_objects._validate_shader(
            "fragment", program["second_fragment"], allow_null=False
        )
        assert isinstance(colormap_fragment, shader_objects.Shader)
        _ = shader_objects.ShaderProgram(colormap_vertex, colormap_fragment)


def test_camera_dict_update(osmesa_fake_amr):
    pos = [0.5, 2.0, 3.0]
    osmesa_fake_amr.scene.camera.set_position(pos)

    cdict = osmesa_fake_amr.scene.camera.dict()
    assert_equal(cdict["position"], pos)

    osmesa_fake_amr.scene.camera.set_position([4.0, 4.0, 4])
    osmesa_fake_amr.scene.camera.update(**cdict)
    assert_equal(osmesa_fake_amr.scene.camera.position, pos)


def test_block_collection_grid_ids():
    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    ds = yt.testing.fake_amr_ds()
    wid = ds.domain_width / 20.0 / 2.0
    c = ds.domain_center
    reg = ds.region(c, c - wid, c + wid)
    rc.add_scene(reg, ("stream", "Density"), no_ghost=True)

    block_coll = rc.scene.data_objects[0]
    gl = block_coll.grid_id_list
    assert len(gl) < len(ds.index.grids)
    grids = block_coll.intersected_grids
    assert len(grids) == len(gl)
    rc.osmesa.OSMesaDestroyContext(rc.context)


def test_manual_scene_graph(image_store):
    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    ds = yt.testing.fake_amr_ds()

    c = TrackballCamera.from_dataset(ds)
    rc.scene = SceneGraph(camera=c)
    rc.scene.data_objects.append(BlockCollection(data_source=ds.all_data()))
    rc.scene.data_objects[-1].add_data(("radius"), no_ghost=True)
    rc.scene.components.append(BlockRendering(data=rc.scene.data_objects[-1]))

    image_store(rc)
    rc.osmesa.OSMesaDestroyContext(rc.context)


def test_block_collection_min_max(image_store):
    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    ds = yt.testing.fake_amr_ds()

    c = TrackballCamera.from_dataset(ds)
    rc.scene = SceneGraph(camera=c)
    rc.scene.data_objects.append(
        BlockCollection(
            data_source=ds.all_data(),
            _compute_min_max=False,
            min_val=0.0,
            max_val=10.0,
        )
    )
    rc.scene.data_objects[-1].add_data(("stream", "Density"), no_ghost=True)
    rc.scene.components.append(BlockRendering(data=rc.scene.data_objects[-1]))

    image_store(rc)
    rc.osmesa.OSMesaDestroyContext(rc.context)



import cmyt
from matplotlib.colors import Colormap

# we want to display colormap name then cmyt for sorting purposes
cmyt_names = [
    f"{cm}.cmyt" for cm in dir(cmyt) if isinstance(getattr(cmyt, cm), Colormap)
]


def validate_cmyt_name(cm):
    if cm.endswith(".cmyt"):
        return "cmyt." + cm.split(".")[0]
    return cm



import numpy as np
import traitlets
from yt.utilities.math_utils import (
    get_lookat_matrix,
    get_perspective_matrix,
    quaternion_to_rotation_matrix,
    rotation_matrix_to_quaternion,
)

from yt_idv.cameras.base_camera import BaseCamera
from yt_idv.utilities.camera_utilities import update_orientation


class TrackballCamera(BaseCamera):
    """

    This class implements a basic "Trackball" or "Arcball" camera control system
    that allows for unconstrained 3D rotations without suffering from Gimbal lock.
    Following Ken Shoemake's original C implementation (Graphics Gems IV, III.1)
    we project mouse movements onto the unit sphere and use quaternions to
    represent the corresponding rotation.

    See also:
    https://en.wikibooks.org/wiki/OpenGL_Programming/Modern_OpenGL_Tutorial_Arcball

    """

    @property
    def proj_func(self):
        return get_perspective_matrix

    @traitlets.default("orientation")
    def _orientation_default(self):
        rotation_matrix = self.view_matrix[0:3, 0:3]
        return rotation_matrix_to_quaternion(rotation_matrix)

    @traitlets.default("view_matrix")
    def _default_view_matrix(self):
        return get_lookat_matrix(self.position, self.focus, self.up)

    def _map_to_surface(self, mouse_x, mouse_y):
        # right now this just maps to the surface of the unit sphere
        x, y = mouse_x, mouse_y
        mag = np.sqrt(x * x + y * y)
        if mag > 1.0:
            x /= mag
            y /= mag
            z = 0.0
        else:
            z = np.sqrt(1.0 - mag**2)
        return np.array([x, -y, z])

    def update_orientation(self, start_x, start_y, end_x, end_y):
        self.orientation = update_orientation(
            self.orientation.astype("float64"), start_x, start_y, end_x, end_y
        )

        rotation_matrix = quaternion_to_rotation_matrix(self.orientation)
        dp = np.linalg.norm(self.position - self.focus) * rotation_matrix[2]
        self.position = dp + self.focus
        self.up = rotation_matrix[1]

        self.view_matrix = get_lookat_matrix(self.position, self.focus, self.up)

        self.projection_matrix = self.proj_func(
            self.fov, self.aspect_ratio, self.near_plane, self.far_plane
        )

    def _update_matrices(self):
        self.view_matrix = get_lookat_matrix(self.position, self.focus, self.up)
        self.orientation = rotation_matrix_to_quaternion(self.view_matrix[0:3, 0:3])

        self.projection_matrix = self.proj_func(
            self.fov, self.aspect_ratio, self.near_plane, self.far_plane
        )

    def move_forward(self, move_amount):
        dpos = (self.focus - self.position) / np.linalg.norm(self.focus - self.position)
        self.offset_position(move_amount * dpos)

    def offset_position(self, dpos=None):
        if dpos is None:
            dpos = np.array([0.0, 0.0, 0.0])
        self.position += dpos
        self.view_matrix = get_lookat_matrix(self.position, self.focus, self.up)

    def _compute_matrices(self):
        pass

    def set_position(self, pos):
        self.position = pos
        self._update_matrices()

    @staticmethod
    def from_dataset(ds):
        center, pos, near_plane = _get_camera_for_ds_geometry(ds)

        c = TrackballCamera(position=pos, focus=center, near_plane=near_plane)
        c.update_orientation(0, 0, 0, 0)
        return c


def _get_camera_for_ds_geometry(ds):

    if str(ds.geometry) == "spherical":
        # dummy values here, will get updated after data is loaded
        # and the cartesian bounds are available
        center = np.array([0.5, 0.5, 0.5])
        wid = np.array([1.0, 1.0, 1.0])
        pos = center + 1.5 * wid
        dx_aprox = wid[0] / np.max(ds.domain_dimensions)
        near_plane = 3.0 * dx_aprox
        near_plane = max(near_plane, 1e-5)
    elif str(ds.geometry) == "cartesian":
        center = ds.domain_center
        pos = center + 1.5 * ds.domain_width.in_units("unitary")
        near_plane = 3.0 * ds.index.get_smallest_dx().min().in_units("unitary").d
        near_plane = max(near_plane, 1e-5)
    else:
        raise NotImplementedError(
            "Only cartesian and spherical geometries are supported at present."
        )
    return center, pos, near_plane



"""
Shader and ShaderProgram wrapper classes for vertex and fragment shaders used
in Interactive Data Visualization
"""

# ----------------------------------------------------------------------------
# Copyright (c) 2016, yt Development Team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file COPYING.txt, distributed with this software.
# ----------------------------------------------------------------------------

# This is a part of the experimental Interactive Data Visualization

from contextlib import ExitStack, contextmanager

import matplotlib.pyplot as plt
import numpy as np
import traitlets
import traittypes
from OpenGL import GL

# Set up a mapping from numbers to names
from yt.utilities.math_utils import get_scale_matrix, get_translate_matrix

from yt_idv._cmyt_utilities import validate_cmyt_name
from yt_idv.constants import bbox_vertices

const_types = (
    GL.constant.IntConstant,
    GL.constant.LongConstant,
    GL.constant.FloatConstant,
)

num_to_const = {}
for i in dir(GL):
    if i.startswith("GL_"):
        v = getattr(GL, i)
        if not isinstance(v, const_types):
            continue
        num_to_const[v.real] = v

_coersion_funcs = {
    "FLOAT": float,
    "DOUBLE": float,
    "INT": int,
    "UNSIGNED": int,
    "BOOL": bool,
}

_shapes = {
    "MAT2": (2, 2),
    "MAT3": (3, 3),
    "MAT4": (4, 4),
    "MAT2x3": (2, 3),
    "MAT2x4": (2, 4),
    "MAT3x2": (3, 2),
    "MAT3x4": (3, 4),
    "MAT4x2": (4, 2),
    "MAT4x3": (4, 3),
    "VEC2": (2,),
    "VEC3": (3,),
    "VEC4": (4,),
}

# PyOpenGL has the reverse mapping for this
gl_to_np = {
    "FLOAT": "f",
    "DOUBLE": "d",
    "INT": "i",
    "UNSIGNED": "I",
    "BOOL": "b",
}

np_to_gl = {
    "float32": GL.GL_FLOAT,
    "uint32": GL.GL_UNSIGNED_INT,
    "uint8": GL.GL_UNSIGNED_BYTE,
}

TEX_CHANNELS = {
    "float32": {
        1: (GL.GL_FLOAT, GL.GL_R32F, GL.GL_RED),
        2: (GL.GL_FLOAT, GL.GL_RG32F, GL.GL_RG),
        3: (GL.GL_FLOAT, GL.GL_RGB32F, GL.GL_RGB),
        4: (GL.GL_FLOAT, GL.GL_RGBA32F, GL.GL_RGBA),
    },
    "uint8": {
        1: (GL.GL_UNSIGNED_BYTE, GL.GL_R8, GL.GL_RED),
        2: (GL.GL_UNSIGNED_BYTE, GL.GL_RG8, GL.GL_RG),
        3: (GL.GL_UNSIGNED_BYTE, GL.GL_RGB8, GL.GL_RGB),
        4: (GL.GL_UNSIGNED_BYTE, GL.GL_RGBA8, GL.GL_RGBA),
    },
    "uint32": {
        1: (GL.GL_UNSIGNED_INT, GL.GL_R32UI, GL.GL_RED),
        2: (GL.GL_UNSIGNED_INT, GL.GL_RG32UI, GL.GL_RG),
        3: (GL.GL_UNSIGNED_INT, GL.GL_RGB32UI, GL.GL_RGB),
        4: (GL.GL_UNSIGNED_INT, GL.GL_RGBA32UI, GL.GL_RGBA),
    },
}


def coerce_uniform_type(val, gl_type):
    # gl_type here must be in const_types
    if not isinstance(gl_type, const_types):
        gl_type = num_to_const[gl_type]
    # Now we can get down to business!
    spec = gl_type.name.split("_")[1:]  # Strip out the GL_
    # We know what to do with:
    #   FLOAT DOUBLE INT UNSIGNED BOOL
    # We can ignore:
    #    SAMPLER IMAGE
    if "SAMPLER" in spec or "IMAGE" in spec:
        # Do nothing to these, and let PyOpenGL handle it
        return val
    if len(spec) == 1 or spec == ["UNSIGNED", "INT"]:
        return _coersion_funcs[spec[0]](val)
    # We need to figure out if it's a matrix, a vector, etc.
    shape = _shapes[spec[-1]]
    dtype = gl_to_np[spec[0]]
    val = np.asanyarray(val, dtype=dtype)
    val.shape = shape
    return val


class TextureBoundary(traitlets.TraitType):
    default_value = GL.GL_CLAMP_TO_EDGE
    info_text = "A boundary type of mirror, clamp, or repeat"

    def validate(self, obj, value):
        if isinstance(value, str):
            try:
                return {
                    "clamp": GL.GL_CLAMP_TO_EDGE,
                    "mirror": GL.GL_MIRRORED_REPEAT,
                    "repeat": GL.GL_REPEAT,
                }[value.lower()]
            except KeyError:
                self.error(obj, value)
        elif value in (GL.GL_CLAMP_TO_EDGE, GL.GL_MIRRORED_REPEAT, GL.GL_REPEAT):
            return value
        self.error(obj, value)


class GLValue(traitlets.TraitType):
    default_value = GL.GL_NONE
    info_text = "An OpenGL constant."

    def validate(self, obj, value):
        # This will convert lower to upper and spaces to _ and also preprend
        # GL_ if needed.
        if isinstance(value, str):
            if not value.startswith("GL"):
                value = f"GL_{value}"
            value = getattr(GL, value.upper().replace(" ", "_"), None)
            if value is None:
                self.error(obj, value)
        return value


TEX_TARGETS = {i: getattr(GL, f"GL_TEXTURE{i}") for i in range(10)}


class Texture(traitlets.HasTraits):
    texture_name = traitlets.CInt(-1)
    data = traittypes.Array(None, allow_none=True)
    channels = GLValue("r32f")
    min_filter = GLValue("linear")
    mag_filter = GLValue("linear")

    @traitlets.default("texture_name")
    def _default_texture_name(self):
        return GL.glGenTextures(1)

    @contextmanager
    def bind(self, target=0):
        _ = GL.glActiveTexture(TEX_TARGETS[target])
        _ = GL.glBindTexture(self.dim_enum, self.texture_name)
        yield
        _ = GL.glActiveTexture(TEX_TARGETS[target])
        GL.glBindTexture(self.dim_enum, 0)


class Texture1D(Texture):
    boundary_x = TextureBoundary()
    dims = 1
    dim_enum = GLValue("texture 1d")

    @traitlets.observe("data")
    def _set_data(self, change):
        with self.bind():
            data = change["new"]
            if len(data.shape) == 2:
                channels = data.shape[-1]
            else:
                channels = 1
            dx = data.shape[0]
            gl_type, type1, type2 = TEX_CHANNELS[data.dtype.name][channels]
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
            if not isinstance(change["old"], np.ndarray):
                GL.glTexStorage1D(GL.GL_TEXTURE_1D, 6, type1, dx)
            GL.glTexSubImage1D(GL.GL_TEXTURE_1D, 0, 0, dx, type2, gl_type, data)
            GL.glTexParameterf(GL.GL_TEXTURE_1D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameteri(
                GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MIN_FILTER, self.min_filter
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_1D, GL.GL_TEXTURE_MAG_FILTER, self.mag_filter
            )
            GL.glGenerateMipmap(GL.GL_TEXTURE_1D)


class ColormapTexture(Texture1D):
    colormap_name = traitlets.CUnicode()

    def __init__(self, *args, **kwargs):
        # Override...
        kwargs["boundary_x"] = "clamp"
        super().__init__(*args, **kwargs)

    @traitlets.validate("colormap_name")
    def _validate_name(self, proposal):
        try:
            plt.get_cmap(validate_cmyt_name(proposal["value"]))
        except ValueError:
            raise traitlets.TraitError(
                "Colormap name needs to be known by" "matplotlib"
            )
        return proposal["value"]

    @traitlets.observe("colormap_name")
    def _observe_colormap_name(self, change):
        cmap = plt.get_cmap(validate_cmyt_name(change["new"]))
        cmap_vals = np.array(cmap(np.linspace(0, 1, 256)), dtype="f4")
        self.data = cmap_vals


class Texture2D(Texture):
    boundary_x = TextureBoundary()
    boundary_y = TextureBoundary()
    dims = 2
    channels = 1
    dim_enum = GLValue("texture 2d")

    @traitlets.observe("data")
    def _set_data(self, change):
        with self.bind():
            data = change["new"]
            if len(data.shape) == 3:
                channels = data.shape[-1]
            else:
                channels = 1
            self.channels = channels
            dx, dy = data.shape[:2]
            gl_type, type1, type2 = TEX_CHANNELS[data.dtype.name][channels]
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
            if not isinstance(change["old"], np.ndarray):
                GL.glTexStorage2D(GL.GL_TEXTURE_2D, 2, type1, dx, dy)
            GL.glTexSubImage2D(
                GL.GL_TEXTURE_2D, 0, 0, 0, dx, dy, type2, gl_type, data.swapaxes(0, 1)
            )
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, self.boundary_y)
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MIN_FILTER, self.min_filter
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_2D, GL.GL_TEXTURE_MAG_FILTER, self.mag_filter
            )
            GL.glGenerateMipmap(GL.GL_TEXTURE_2D)


class TransferFunctionTexture(Texture2D):
    def __init__(self, *args, **kwargs):
        kwargs["boundary_x"] = "clamp"
        kwargs["boundary_y"] = "clamp"
        super().__init__(*args, **kwargs)


class DepthBuffer(Texture2D):
    def create_texture(self, w, h):
        with self.bind():
            GL.glTexImage2D(
                GL.GL_TEXTURE_2D,
                0,
                GL.GL_DEPTH_COMPONENT24,
                w,
                h,
                0,
                GL.GL_DEPTH_COMPONENT,
                GL.GL_FLOAT,
                None,
            )
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameterf(GL.GL_TEXTURE_2D, GL.GL_TEXTURE_WRAP_T, self.boundary_y)
            GL.glGenerateMipmap(GL.GL_TEXTURE_2D)


class Texture3D(Texture):
    boundary_x = TextureBoundary()
    boundary_y = TextureBoundary()
    boundary_z = TextureBoundary()
    dims = 3
    dim_enum = GLValue("texture 3d")

    @traitlets.observe("data")
    def _set_data(self, change):
        with self.bind():
            data = change["new"]
            if len(data.shape) == 4:
                channels = data.shape[-1]
            else:
                channels = 1
            dx, dy, dz = data.shape[:3]
            gl_type, type1, type2 = TEX_CHANNELS[data.dtype.name][channels]
            GL.glPixelStorei(GL.GL_UNPACK_ALIGNMENT, 1)
            GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_S, self.boundary_x)
            GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_T, self.boundary_y)
            GL.glTexParameterf(GL.GL_TEXTURE_3D, GL.GL_TEXTURE_WRAP_R, self.boundary_z)
            if not isinstance(change["old"], np.ndarray):
                GL.glTexStorage3D(GL.GL_TEXTURE_3D, 1, type1, dx, dy, dz)
            GL.glTexSubImage3D(
                GL.GL_TEXTURE_3D, 0, 0, 0, 0, dx, dy, dz, type2, gl_type, data.T
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MIN_FILTER, self.min_filter
            )
            GL.glTexParameteri(
                GL.GL_TEXTURE_3D, GL.GL_TEXTURE_MAG_FILTER, self.mag_filter
            )
            GL.glGenerateMipmap(GL.GL_TEXTURE_3D)


class VertexAttribute(traitlets.HasTraits):
    name = traitlets.CUnicode("attr")
    id = traitlets.CInt(-1)
    data = traittypes.Array(None, allow_none=True)
    each = traitlets.CInt(-1)
    opengl_type = traitlets.CInt(GL.GL_FLOAT)
    divisor = traitlets.CInt(0)

    @traitlets.default("id")
    def _id_default(self):
        return GL.glGenBuffers(1)

    @contextmanager
    def bind(self, program=None):
        loc = -1
        if program is not None:
            loc = GL.glGetAttribLocation(program.program, self.name)
            if loc >= 0:
                GL.glVertexAttribDivisor(loc, self.divisor)
                _ = GL.glEnableVertexAttribArray(loc)
        _ = GL.glBindBuffer(GL.GL_ARRAY_BUFFER, self.id)
        if loc >= 0:
            GL.glVertexAttribPointer(loc, self.each, self.opengl_type, False, 0, None)
        yield
        if loc >= 0:
            GL.glDisableVertexAttribArray(loc)
        GL.glBindBuffer(GL.GL_ARRAY_BUFFER, 0)

    @traitlets.observe("data")
    def _set_data(self, change):
        arr = change["new"]
        self.each = arr.shape[-1]
        self.opengl_type = np_to_gl[arr.dtype.name]
        with self.bind():
            GL.glBufferData(GL.GL_ARRAY_BUFFER, arr.nbytes, arr, GL.GL_STATIC_DRAW)


class VertexArray(traitlets.HasTraits):
    name = traitlets.CUnicode("vertex")
    id = traitlets.CInt(-1)
    indices = traittypes.Array(None, allow_none=True)
    index_id = traitlets.CInt(-1)
    attributes = traitlets.List(trait=traitlets.Instance(VertexAttribute))
    each = traitlets.CInt(-1)

    @traitlets.default("id")
    def _id_default(self):
        return GL.glGenVertexArrays(1)

    @contextmanager
    def bind(self, program=None):
        GL.glBindVertexArray(self.id)
        if self.index_id != -1:
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.index_id)
        # We only bind the attributes if we have a program too
        if program is None:
            attrs = []
        else:
            attrs = self.attributes
        with ExitStack() as stack:
            _ = [stack.enter_context(_.bind(program)) for _ in attrs]
            yield
        if self.index_id != -1:
            GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, 0)
        GL.glBindVertexArray(0)

    @traitlets.observe("indices")
    def _set_indices(self, change):
        arr = change["new"]
        self.index_id = GL.glGenBuffers(1)
        GL.glBindBuffer(GL.GL_ELEMENT_ARRAY_BUFFER, self.index_id)
        GL.glBufferData(GL.GL_ELEMENT_ARRAY_BUFFER, arr.nbytes, arr, GL.GL_STATIC_DRAW)


class Framebuffer(traitlets.HasTraits):
    fb_id = traitlets.CInt(-1)
    rb_id = traitlets.CInt(-1)
    fb_tex = traitlets.Instance(Texture2D)
    db_tex = traitlets.Instance(DepthBuffer)
    viewport = traitlets.Tuple(
        traitlets.CInt(), traitlets.CInt(), traitlets.CInt(), traitlets.CInt()
    )
    initialized = traitlets.Bool(False)

    @property
    def data(self):
        origin_x, origin_y, width, height = self.viewport
        with self.bind(clear=False):
            arr = GL.glReadPixels(0, 0, width, height, GL.GL_RGBA, GL.GL_FLOAT)
        return arr

    @property
    def depth_data(self):
        origin_x, origin_y, width, height = self.viewport
        with self.bind(clear=False):
            arr = GL.glReadPixels(
                0, 0, width, height, GL.GL_DEPTH_COMPONENT, GL.GL_FLOAT
            )
        return arr

    @traitlets.default("viewport")
    def _viewport_default(self):
        # origin_x, origin_y, width, height
        return tuple(GL.glGetIntegerv(GL.GL_VIEWPORT))

    @traitlets.observe("viewport")
    def _viewport_changed(self, change):
        # we just need to disable the initialized value here
        self.initialized = False

    @traitlets.default("fb_id")
    def _fb_id_default(self):
        return GL.glGenFramebuffers(1)

    @traitlets.default("rb_id")
    def _rb_id_default(self):
        return GL.glGenRenderbuffers(1)

    @traitlets.default("fb_tex")
    def _fb_tex_default(self):
        data = np.zeros((self.viewport[2], self.viewport[3], 4), "f4")
        return Texture2D(data=data, boundary_x="repeat", boundary_y="repeat")

    @traitlets.default("db_tex")
    def _db_tex_default(self):
        db = DepthBuffer(boundary_x="repeat", boundary_y="repeat")
        db.create_texture(self.viewport[2], self.viewport[3])
        return db

    @contextmanager
    def bind(self, clear=True):
        self.viewport = tuple(GL.glGetIntegerv(GL.GL_VIEWPORT))
        if not self.initialized:
            GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fb_id)
            GL.glBindRenderbuffer(GL.GL_RENDERBUFFER, self.rb_id)
            GL.glRenderbufferStorage(
                GL.GL_RENDERBUFFER,
                GL.GL_DEPTH_COMPONENT32F,
                self.viewport[2],
                self.viewport[3],
            )
            GL.glFramebufferRenderbuffer(
                GL.GL_FRAMEBUFFER,
                GL.GL_DEPTH_ATTACHMENT,
                GL.GL_RENDERBUFFER,
                self.rb_id,
            )

            GL.glFramebufferTexture2D(
                GL.GL_FRAMEBUFFER,
                GL.GL_COLOR_ATTACHMENT0,
                GL.GL_TEXTURE_2D,
                self.fb_tex.texture_name,
                0,  # mipmap level, normally 0
            )
            GL.glFramebufferTexture2D(
                GL.GL_FRAMEBUFFER,
                GL.GL_DEPTH_ATTACHMENT,
                GL.GL_TEXTURE_2D,
                self.db_tex.texture_name,
                0,  # mipmap level, normally 0
            )
            status = GL.glCheckFramebufferStatus(GL.GL_FRAMEBUFFER)
            if status != GL.GL_FRAMEBUFFER_COMPLETE:
                raise RuntimeError
            self.initialized = True
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, self.fb_id)
        if clear:
            GL.glClear(GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT)
        yield
        GL.glBindFramebuffer(GL.GL_FRAMEBUFFER, 0)

    @contextmanager
    def input_bind(self, fb_target=0, db_target=1):
        with self.fb_tex.bind(fb_target):
            with self.db_tex.bind(db_target):
                yield


class Texture3DIterator(traitlets.HasTraits):
    items = traitlets.Any()

    def __iter__(self, target=0):
        tex_target = TEX_TARGETS[target]
        for i, t in self.items:
            GL.glActiveTexture(tex_target)
            GL.glBindTexture(GL.GL_TEXTURE_3D, t.texture_name)
            yield i
        GL.glActiveTexture(tex_target)
        GL.glBindTexture(GL.GL_TEXTURE_3D, 0)


def compute_box_geometry(left_edge, right_edge):
    move = get_translate_matrix(*left_edge)
    width = right_edge - left_edge
    scale = get_scale_matrix(*width)

    transformed_box = bbox_vertices.dot(scale.T).dot(move.T).astype("float32")
    return transformed_box



import argparse

import numpy as np
import yt

import yt_idv

# yt reminder: phi is the azimuthal angle (0 to 2pi)
# theta is the co-latitude, the angle from north (0 to pi)
# coord ordering here will be r, phi, theta
bbox_options = {
    "partial": np.array([[0.5, 1.0], [0.0, np.pi / 3], [np.pi / 4, np.pi / 2]]),
    "whole": np.array([[0.0, 1.0], [0.0, 2 * np.pi], [0, np.pi]]),
    "shell": np.array([[0.7, 1.0], [0.0, 2 * np.pi], [0, np.pi]]),
    "north_hemi": np.array([[0.1, 1.0], [0.0, 2 * np.pi], [0, 0.5 * np.pi]]),
    "north_shell": np.array([[0.8, 1.0], [0.0, 2 * np.pi], [0, 0.5 * np.pi]]),
    "south_hemi": np.array([[0.1, 1.0], [0.0, 2 * np.pi], [0.5 * np.pi, np.pi]]),
    "ew_hemi": np.array([[0.1, 1.0], [0.0, np.pi], [0.0, np.pi]]),
    "quadrant_shell": np.array([[0.6, 1.0], [0.0, np.pi / 2], [0.0, np.pi / 2]]),
}

if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog="spherical_amr_rendering",
        description="Loads an example spherical dataset in yt_idv",
    )

    msg = f"The geometry subset to generate: one of {list(bbox_options.keys())}"
    parser.add_argument("-d", "--domain", default="partial", help=msg)
    msg = (
        "The field to plot. Provide a comma-separated string with field_type,field "
        "e.g., to plot the field tuple ('index', 'phi'): \n "
        "    $ python amr_spherical_volume_rendering.py -f index,x "
        "\nIf a single string is provided, a field type of gas is assumed."
    )
    parser.add_argument("-f", "--field", default="index,phi", help=msg)
    parser.add_argument(
        "-np", "--nprocs", default=64, help="number of grids to decompose domain to"
    )
    parser.add_argument(
        "-sz", "--size", default=256, help="dimensions, will be (size, size size)"
    )

    args = parser.parse_args()

    sz = (int(args.size),) * 3
    fake_data = {"density": np.random.random(sz)}

    field = str(args.field).split(",")
    if len(field) == 1:
        field = ("gas", str(field).strip())
    elif len(field) == 2:
        field = (field[0].strip(), field[1].strip())
    else:
        raise RuntimeError(
            "Unexpected field formatting. Provide a single string"
            " to provide just a field (will assume field type "
            " of 'gas', or a comma separated string to provide a "
            "field type and a field"
        )

    if args.domain not in bbox_options:
        raise RuntimeError(
            f"domain must be one of {list(bbox_options.keys())}, found {args.domain}"
        )
    bbox = bbox_options[args.domain]

    nprocs = int(args.nprocs)

    ds = yt.load_uniform_grid(
        fake_data,
        sz,
        bbox=bbox,
        nprocs=nprocs,
        geometry="spherical",
        axis_order=("r", "phi", "theta"),
        length_unit=1,
    )

    phi_c = ds.quan(ds.domain_center[ds.coordinates.axis_id["phi"]].d, "")
    theta_c = ds.quan(ds.domain_center[ds.coordinates.axis_id["theta"]].d, "")
    rmax = ds.domain_right_edge[ds.coordinates.axis_id["r"]]
    phi_f = ds.quan(15.0 * np.pi / 180.0, "")
    theta_f = ds.quan(15.0 * np.pi / 180.0, "")
    min_val = ds.quan(0.1, "")

    def _tube(field, data):
        phi = data["index", "phi"]
        theta = data["index", "theta"]
        tube = (1 - min_val) * np.exp(-(((theta - theta_c) / theta_f) ** 2))
        tube = tube * np.exp(-(((phi - phi_c) / phi_f) ** 2))
        return tube + min_val

    ds.add_field(
        name=("stream", "tube"),
        function=_tube,
        sampling_type="local",
    )

    def _r_rev(field, data):
        r = data["index", "r"]
        return rmax - r

    ds.add_field(
        name=("stream", "r_rev"),
        function=_r_rev,
        sampling_type="local",
    )

    if field not in ds.field_list + ds.derived_field_list:
        spaces = " " * 8
        fld_list_str = f"\n{spaces}".join(str(fld) for fld in ds.field_list)
        drv_fld_list_str = f"\n{spaces}".join(str(fld) for fld in ds.derived_field_list)
        raise RuntimeError(
            f"field {field} not in field_list or derived_field_list:\n"
            f"\n    ds.field_list:\n{spaces}{fld_list_str}"
            f"\n    ds.derived_field_list:\n{spaces}{drv_fld_list_str}"
        )

    rc = yt_idv.render_context(height=800, width=800, gui=True)
    sg = rc.add_scene(ds, field, no_ghost=True)
    rc.scene.components[0].sample_factor = 5.0
    rc.scene.components[0].cmap_log = False
    rc.scene.components[0]._reset_cmap_bounds()

    rc.run()



import numpy as np
import traitlets
from yt.data_objects.data_containers import YTDataContainer
from yt.utilities.lib.mesh_triangulation import triangulate_mesh

from yt_idv.opengl_support import Texture3D, VertexArray, VertexAttribute
from yt_idv.scene_data.base_data import SceneData


class MeshData(SceneData):
    name = "mesh"
    data_source = traitlets.Instance(YTDataContainer)
    texture_objects = traitlets.Dict(trait=traitlets.Instance(Texture3D))
    texture_objects = traitlets.Dict(trait=traitlets.Instance(Texture3D))
    blocks = traitlets.Dict(default_value=())
    scale = traitlets.Bool(False)
    size = traitlets.CInt(-1)

    def get_mesh_data(self, data_source, field):
        """

        This reads the mesh data into a form that can be fed in to OpenGL.

        """

        # get mesh information
        try:
            ftype, fname = field
            mesh_id = int(ftype[-1])
        except ValueError:
            mesh_id = 0

        mesh = data_source.ds.index.meshes[mesh_id - 1]
        offset = mesh._index_offset
        vertices = mesh.connectivity_coords
        if hasattr(vertices, "in_units"):
            vertices = vertices.in_units("unitary")
        indices = mesh.connectivity_indices - offset

        data = data_source[field]

        return triangulate_mesh(vertices, data, indices)

    def add_data(self, field):
        v, d, i = self.get_mesh_data(self.data_source, field)
        v.shape = (v.size // 3, 3)
        v = np.concatenate([v, np.ones((v.shape[0], 1))], axis=-1).astype("f4")
        d.shape = (d.size, 1)
        i.shape = (i.size, 1)
        i = i.astype("uint32")
        # d[:] = np.mgrid[0.0:1.0:1j*d.size].astype("f4")[:,None]
        self.vertex_array.attributes.append(
            VertexAttribute(name="model_vertex", data=v)
        )
        self.vertex_array.attributes.append(
            VertexAttribute(name="vertex_data", data=d.astype("f4"))
        )
        self.vertex_array.indices = i
        self.size = i.size

    @traitlets.default("vertex_array")
    def _default_vertex_array(self):
        return VertexArray(name="mesh_info", each=0)



import numpy as np
import pytest

from yt_idv.utilities.coordinate_utilities import (
    SphericalMixedCoordBBox,
    cartesian_bboxes,
    cartesian_bboxes_edges,
    cartesian_to_spherical,
    spherical_to_cartesian,
)


def test_cartesian_bboxes_for_spherical():

    # this test checks the special cases where
    # a spherical volume element crosses an axis
    # or when an element edge is lined up with an axis

    # check element that includes theta=0 as an edge
    r = np.array([0.95])
    dr = np.array([0.1])
    theta = np.array([0.05])
    dtheta = np.array([0.1])
    phi = np.array([0.05])
    dphi = np.array([0.05])

    bbox_handler = SphericalMixedCoordBBox()

    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)
    assert xyz[2] + dxyz[2] / 2 == 1.0
    assert np.allclose(xyz[0] - dxyz[0] / 2, 0.0)
    assert np.allclose(xyz[1] - dxyz[1] / 2, 0.0)

    # now theta = np.pi
    theta = np.array([np.pi - dtheta[0] / 2])
    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)
    assert xyz[2] - dxyz[2] / 2 == -1.0
    assert np.allclose(xyz[0] - dxyz[0] / 2, 0.0)
    assert np.allclose(xyz[1] - dxyz[1] / 2, 0.0)

    # element at equator, overlapping the +y axis
    theta = np.array([np.pi / 2])
    phi = np.array([np.pi / 2])
    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)

    assert xyz[1] + dxyz[1] / 2 == 1.0
    assert np.allclose(xyz[0], 0.0)
    assert np.allclose(xyz[2], 0.0)

    # element at equator, overlapping the -x axis
    phi = np.array([np.pi])
    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)

    assert xyz[0] - dxyz[0] / 2 == -1.0
    assert np.allclose(xyz[1], 0.0)
    assert np.allclose(xyz[2], 0.0)

    # element at equator, overlapping the -y axis
    phi = np.array([3 * np.pi / 2])
    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)

    assert xyz[1] - dxyz[1] / 2 == -1.0
    assert np.allclose(xyz[0], 0.0)
    assert np.allclose(xyz[2], 0.0)

    # element at equator, overlapping +x axis
    phi = dphi / 2.0
    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)
    assert xyz[0] + dxyz[0] / 2 == 1.0

    # element with edge on +x axis in -theta direction
    theta = np.pi / 2 - dtheta / 2
    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)
    assert xyz[0] + dxyz[0] / 2 == 1.0

    # element with edge on +x axis in +theta direction
    theta = np.pi / 2 + dtheta / 2
    xyz, dxyz = cartesian_bboxes(bbox_handler, r, theta, phi, dr, dtheta, dphi)
    assert xyz[0] + dxyz[0] / 2 == 1.0

    # finally, check that things work OK with a wide range of
    # angles

    r_edges = np.linspace(0.4, 1.0, 10, dtype="float64")
    theta_edges = np.linspace(0, np.pi, 10, dtype="float64")
    phi_edges = np.linspace(0.0, 2 * np.pi, 10, dtype="float64")

    r = (r_edges[0:-1] + r_edges[1:]) / 2.0
    theta = (theta_edges[0:-1] + theta_edges[1:]) / 2.0
    phi = (phi_edges[0:-1] + phi_edges[1:]) / 2.0

    dr = r_edges[1:] - r_edges[:-1]
    dtheta = theta_edges[1:] - theta_edges[:-1]
    dphi = phi_edges[1:] - phi_edges[:-1]

    r_th_ph = np.meshgrid(r, theta, phi)
    d_r_th_ph = np.meshgrid(dr, dtheta, dphi)
    r_th_ph = [r_th_ph[i].ravel() for i in range(3)]
    d_r_th_ph = [d_r_th_ph[i].ravel() for i in range(3)]

    x_y_z, d_x_y_z = cartesian_bboxes(
        bbox_handler,
        r_th_ph[0],
        r_th_ph[1],
        r_th_ph[2],
        d_r_th_ph[0],
        d_r_th_ph[1],
        d_r_th_ph[2],
    )

    assert np.all(np.isfinite(x_y_z))
    assert np.all(np.isfinite(d_x_y_z))

    # and check the extents again for completeness...
    for i in range(3):
        max_val = np.max(x_y_z[i] + d_x_y_z[i] / 2.0)
        min_val = np.min(x_y_z[i] - d_x_y_z[i] / 2.0)
        assert max_val == 1.0
        assert min_val == -1.0


def test_spherical_cartesian_roundtrip():
    xyz = [np.linspace(0, 1, 10) for _ in range(3)]
    xyz = np.meshgrid(*xyz)
    xyz = [xyzi.ravel() for xyzi in xyz]
    x, y, z = xyz

    r, theta, phi = cartesian_to_spherical(x, y, z)
    x_out, y_out, z_out = spherical_to_cartesian(r, theta, phi)

    assert np.allclose(x_out, x)
    assert np.allclose(y_out, y)
    assert np.allclose(z_out, z)
    assert np.max(r) == np.sqrt(3.0)


@pytest.mark.parametrize("n_angles", (2, 4, 8, 16))
def test_large_elements(n_angles):

    bbox_handler = SphericalMixedCoordBBox()

    r_edges = np.linspace(0.4, 1.0, 10, dtype="float64")
    theta_edges = np.linspace(0, np.pi, n_angles, dtype="float64")
    phi_edges = np.linspace(0.0, 2 * np.pi, n_angles, dtype="float64")

    r = (r_edges[0:-1] + r_edges[1:]) / 2.0
    theta = (theta_edges[0:-1] + theta_edges[1:]) / 2.0
    phi = (phi_edges[0:-1] + phi_edges[1:]) / 2.0

    dr = r_edges[1:] - r_edges[:-1]
    dtheta = theta_edges[1:] - theta_edges[:-1]
    dphi = phi_edges[1:] - phi_edges[:-1]

    r_th_ph = np.meshgrid(r, theta, phi)
    d_r_th_ph = np.meshgrid(dr, dtheta, dphi)
    r_th_ph = [r_th_ph[i].ravel() for i in range(3)]
    d_r_th_ph = [d_r_th_ph[i].ravel() for i in range(3)]

    x_y_z, d_x_y_z = cartesian_bboxes(
        bbox_handler,
        r_th_ph[0],
        r_th_ph[1],
        r_th_ph[2],
        d_r_th_ph[0],
        d_r_th_ph[1],
        d_r_th_ph[2],
    )

    x_y_z = np.column_stack(x_y_z)
    d_x_y_z = np.column_stack(d_x_y_z)

    assert np.all(np.isfinite(x_y_z))
    assert np.all(np.isfinite(d_x_y_z))

    le = np.min(x_y_z - d_x_y_z / 2.0, axis=0)
    re = np.max(x_y_z + d_x_y_z / 2.0, axis=0)

    assert np.all(le == -1.0)
    assert np.all(re == 1.0)


def test_spherical_boxes_edges():
    # check that you get the same result from supplying centers
    # vs edges.

    bbox_handler = SphericalMixedCoordBBox()

    r_edges = np.linspace(0.0, 10.0, 20, dtype="float64")
    theta_edges = np.linspace(0, np.pi, 20, dtype="float64")
    phi_edges = np.linspace(0.0, 2 * np.pi, 20, dtype="float64")

    r = (r_edges[0:-1] + r_edges[1:]) / 2.0
    theta = (theta_edges[0:-1] + theta_edges[1:]) / 2.0
    phi = (phi_edges[0:-1] + phi_edges[1:]) / 2.0

    dr = r_edges[1:] - r_edges[:-1]
    dtheta = theta_edges[1:] - theta_edges[:-1]
    dphi = phi_edges[1:] - phi_edges[:-1]

    r_th_ph = np.meshgrid(r, theta, phi)
    d_r_th_ph = np.meshgrid(dr, dtheta, dphi)
    r_th_ph = [r_th_ph[i].ravel() for i in range(3)]
    d_r_th_ph = [d_r_th_ph[i].ravel() for i in range(3)]

    x_y_z, d_x_y_z = cartesian_bboxes(
        bbox_handler,
        r_th_ph[0],
        r_th_ph[1],
        r_th_ph[2],
        d_r_th_ph[0],
        d_r_th_ph[1],
        d_r_th_ph[2],
    )
    x_y_z = np.column_stack(x_y_z)
    d_x_y_z = np.column_stack(d_x_y_z)

    le = [r_th_ph[i] - d_r_th_ph[i] / 2.0 for i in range(3)]
    re = [r_th_ph[i] + d_r_th_ph[i] / 2.0 for i in range(3)]
    xyz_le, xyz_re = cartesian_bboxes_edges(
        bbox_handler,
        le[0],
        le[1],
        le[2],
        re[0],
        re[1],
        re[2],
    )
    xyz_le = np.column_stack(xyz_le)
    xyz_re = np.column_stack(xyz_re)

    centers = (xyz_le + xyz_re) / 2.0
    widths = xyz_re - xyz_le

    assert np.allclose(centers, x_y_z)
    assert np.allclose(widths, d_x_y_z)



import numpy as np
import yt
from yt.units import Mpc
from yt.visualization.api import Streamlines

import yt_idv
from yt_idv.scene_components.curves import (  # NOQA
    CurveCollectionRendering,
    CurveRendering,
)
from yt_idv.scene_data.curve import CurveCollection, CurveData  # NOQA

ds = yt.load_sample("IsolatedGalaxy")

# generate some streamlines
# (see https://yt-project.org/doc/visualizing/streamlines.html)
c = ds.domain_center  # the center of the box
N = 10  # number of streamlines
scale = ds.domain_width[0]  # scale of streamlines relative to boxsize
pos_dx = np.random.random((N, 3)) * scale - scale / 2.0
pos = c + pos_dx  # starting positions

streamlines = Streamlines(
    ds,
    pos,
    ("gas", "velocity_x"),
    ("gas", "velocity_y"),
    ("gas", "velocity_z"),
    length=1.0 * Mpc,
    get_magnitude=True,
)

streamlines.integrate_through_volume()

# initialize the rendering context and scene
rc = yt_idv.render_context(height=800, width=800, gui=True)
sg = rc.add_scene(ds, "density", no_ghost=True)


# add a single streamline as a single curve
curved = CurveData()
curved.add_data(streamlines.streamlines[0])
curve_render = CurveRendering(data=curved, curve_rgba=(1.0, 0.0, 0.0, 1.0))
curve_render.display_name = "single streamline"
rc.scene.data_objects.append(curved)
rc.scene.components.append(curve_render)

# add the remaining streamlines as a collection of curves
curve_collection = CurveCollection()
for stream in streamlines.streamlines[1:]:
    stream = stream[np.all(stream != 0.0, axis=1)]
    curve_collection.add_curve(stream)
curve_collection.add_data()  # call add_data() after done adding curves

cc_render = CurveCollectionRendering(data=curve_collection)
cc_render.display_name = "multiple streamlines"
rc.scene.data_objects.append(curve_collection)
rc.scene.components.append(cc_render)

rc.run()



import yt

import yt_idv

ds = yt.load_sample("IsolatedGalaxy")
dd = ds.all_data()

rc = yt_idv.render_context("egl", width=1024, height=1024)
rc.add_scene(dd, "density", no_ghost=True)
rc.scene.components[0].visible = False

from yt_idv.scene_annotations.grid_outlines import GridOutlines  # NOQA
from yt_idv.scene_data.grid_positions import GridPositions  # NOQA

grids = ds.index.grids.tolist()

gp = GridPositions(grid_list=grids)
rc.scene.data_objects.append(gp)
go = GridOutlines(data=gp)
rc.scene.components.append(go)

image = rc.run()
yt.write_bitmap(image, "grid_outline.png")



import math

import numpy as np
import traitlets
from OpenGL import GL
from yt.utilities.lib.pixelization_routines import SPHKernelInterpolationTable

from yt_idv.scene_components.base_component import SceneComponent
from yt_idv.scene_data.particle_positions import ParticlePositions

from ..opengl_support import Texture1D

KERNEL_TYPES = [
    "cubic",
    "quartic",
    "quintic",
    "wendland2",
    "wendland4",
    "wendland6",
    "flat",
]


class SPHRendering(SceneComponent):
    name = "sph_rendering"
    data = traitlets.Instance(ParticlePositions)
    scale = traitlets.CFloat(1.0)
    max_particle_size = traitlets.CFloat(1e-3)
    kernel_name = traitlets.Unicode("cubic")
    kernel_table = traitlets.Instance(SPHKernelInterpolationTable, allow_none=True)
    interpolation_texture = traitlets.Instance(Texture1D)

    @traitlets.observe("kernel_name")
    def _change_kernel_name(self, change):
        if change["new"] == "flat":
            self.kernel_table = None
            flat = np.ones(256, dtype="f4")
            flat[-1] = 0.0
            self.interpolation_texture = Texture1D(data=flat)
            return
        self.kernel_table = SPHKernelInterpolationTable(change["new"])
        values = self.kernel_table.interpolate_array(np.mgrid[0.0:1.0:256j])
        self.interpolation_texture = Texture1D(data=values.astype("f4"))

    @traitlets.default("kernel_table")
    def _default_kernel_table(self):
        if self.kernel_name is None:
            return
        return SPHKernelInterpolationTable(self.kernel_name)

    @traitlets.default("interpolation_texture")
    def _default_interpolation_texture(self):
        if self.kernel_table is None and self.kernel_name == "flat":
            flat = np.ones(256, dtype="f4")
            flat[-1] = 0.0
            return Texture1D(data=flat)
        values = self.kernel_table.interpolate_array(np.mgrid[0.0:1.0:256j])
        return Texture1D(data=values.astype("f4"))

    def render_gui(self, imgui, renderer, scene):
        changed = super().render_gui(imgui, renderer, scene)
        _, kernel_index = imgui.listbox(
            "Kernel", KERNEL_TYPES.index(self.kernel_name), KERNEL_TYPES
        )
        if _:
            self.kernel_name = KERNEL_TYPES[kernel_index]
        changed = changed or _
        _, new_value = imgui.slider_float(
            "Log Scale", math.log10(self.scale), -8.0, 2.0
        )
        if _:
            self.scale = 10**new_value
            changed = True
        imgui.text("Filter Particle Max Size")
        _, new_value = imgui.slider_float(
            "max-size", 1.0 / self.max_particle_size, 1.0, 100.0
        )
        if _:
            self.max_particle_size = 1.0 / new_value
            changed = True
        return changed

    def draw(self, scene, program):
        GL.glEnable(GL.GL_CULL_FACE)
        GL.glCullFace(GL.GL_BACK)
        with self.interpolation_texture.bind(0):
            GL.glDrawArraysInstanced(GL.GL_TRIANGLE_STRIP, 0, 4, self.data.size)

    def _set_uniforms(self, scene, shader_program):
        cam = scene.camera
        shader_program._set_uniform("scale", self.scale)
        shader_program._set_uniform("projection", cam.projection_matrix)
        shader_program._set_uniform("modelview", cam.view_matrix)
        shader_program._set_uniform("max_particle_size", self.max_particle_size)
        shader_program._set_uniform(
            "inv_pmvm", np.linalg.inv(cam.projection_matrix @ cam.view_matrix)
        )



def add_popup_help(imgui, message: str) -> bool:
    # adds a hover-triggered popup message to the last-defined imgui item
    if imgui.is_item_hovered():
        imgui.begin_tooltip()
        imgui.text(message)
        imgui.end_tooltip()
        return True
    return False



import base64
import os

import pytest
import yt
from pytest_html import extras as html_extras

import yt_idv


def pytest_configure(config):
    # this will get run before all tests, before collection and
    # any opengl imports that happen within test files.
    os.environ["PYOPENGL_PLATFORM"] = "osmesa"


@pytest.fixture()
def image_store(request, extras, tmpdir):
    def _snap_image(rc):
        image = rc.run()
        img = yt.write_bitmap(image, None)
        content = base64.b64encode(img).decode("ascii")
        extras.append(html_extras.png(content))
        extras.append(html_extras.html("<br clear='all'/>"))

    return _snap_image


@pytest.fixture()
def osmesa_empty_rc():
    """yield an OSMesa empy context then destroy"""

    rc = yt_idv.render_context("osmesa", width=1024, height=1024)
    yield rc
    rc.osmesa.OSMesaDestroyContext(rc.context)
